import functools
from typing import assert_type, assert_never

import attrs
import numpy as np

import caqtus_parsing.nodes as nodes
from caqtus.shot_compilation.timed_instructions import (
    TimedInstruction,
    Pattern,
    create_ramp,
    Concatenated,
    Repeated,
    Ramp,
    merge_instructions,
    concatenate,
)
from caqtus.shot_compilation.timing import Time, number_ticks, start_tick, stop_tick
from caqtus.types.parameter import Parameters
from caqtus.types.units import (
    BaseUnit,
    dimensionless,
    Quantity,
    SECOND,
    Unit,
    InvalidDimensionalityError,
)
from ._is_time_dependent import is_time_dependent
from .._evaluate_scalar_expression import evaluate_expression


@attrs.define
class AnalogInstruction:
    magnitudes: TimedInstruction[np.float64]
    units: BaseUnit


def evaluate_analog_ast(
    ast: nodes.Expression, parameters: Parameters, t1: Time, t2: Time, timestep: Time
) -> AnalogInstruction:
    if not is_time_dependent(ast):
        value = evaluate_expression(ast, parameters)
        length = number_ticks(t1, t2, timestep)
        if isinstance(value, (bool, int, float)):
            return AnalogInstruction(
                magnitudes=Pattern([float(value)]) * length,
                units=dimensionless,
            )
        else:
            assert_type(value, Quantity[float])
            in_base_units = value.to_base_units()
            return AnalogInstruction(
                magnitudes=Pattern([in_base_units.magnitude]) * length,
                units=in_base_units.units,
            )
    else:
        match ast:
            case int() | float() | nodes.Quantity():
                raise AssertionError("Unreachable")
            case nodes.Variable(name=name):
                if name != "t":
                    raise AssertionError("Unreachable")
                tick_start = start_tick(t1, timestep)
                tick_stop = stop_tick(t2, timestep)
                value_start = tick_start * timestep - t1
                value_stop = tick_stop * timestep - t1
                length = tick_stop - tick_start
                return AnalogInstruction(
                    magnitudes=create_ramp(value_start, value_stop, length),
                    units=SECOND,
                )
            case nodes.Call():
                raise NotImplementedError(
                    "Time dependent function calls are not yet supported."
                )
            case nodes.Plus() | nodes.Minus() as unary_operator:
                return evaluate_unary_operator(
                    unary_operator, parameters, t1, t2, timestep
                )
            case nodes.Multiply():
                left = evaluate_analog_ast(ast.left, parameters, t1, t2, timestep)
                right = evaluate_analog_ast(ast.right, parameters, t1, t2, timestep)
                units = left.units * right.units
                assert isinstance(units, Unit)
                return AnalogInstruction(
                    magnitudes=multiply(left.magnitudes, right.magnitudes),
                    units=units.to_base(),
                )
            case nodes.Divide():
                left = evaluate_analog_ast(ast.left, parameters, t1, t2, timestep)
                right = evaluate_analog_ast(ast.right, parameters, t1, t2, timestep)
                units = left.units / right.units
                assert isinstance(units, Unit)
                return AnalogInstruction(
                    magnitudes=divide(left.magnitudes, right.magnitudes),
                    units=units.to_base(),
                )
            case nodes.Add() | nodes.Subtract() as binary_operator:
                left = evaluate_analog_ast(ast.left, parameters, t1, t2, timestep)
                right = evaluate_analog_ast(ast.right, parameters, t1, t2, timestep)
                if left.units != right.units:
                    raise InvalidDimensionalityError(
                        f"Cannot add {ast.left} with units {left.units} to "
                        f"{ast.right} with units {right.units}."
                    )
                left_magnitudes = left.magnitudes
                if isinstance(binary_operator, nodes.Subtract):
                    right_magnitudes = negate(right.magnitudes)
                else:
                    right_magnitudes = right.magnitudes
                return AnalogInstruction(
                    magnitudes=add(left_magnitudes, right_magnitudes),
                    units=left.units,
                )
            case nodes.Power():
                raise NotImplementedError(
                    "Power operator is not yet supported for time dependent "
                    "expressions."
                )
            case _:
                assert_never(ast)


def evaluate_unary_operator(
    unary_operator: nodes.UnaryOperator,
    parameters: Parameters,
    t1: Time,
    t2: Time,
    timestep: Time,
) -> AnalogInstruction:
    operand = evaluate_analog_ast(unary_operator.operand, parameters, t1, t2, timestep)
    match unary_operator:
        case nodes.Plus():
            return operand
        case nodes.Minus():
            return AnalogInstruction(
                magnitudes=negate(operand.magnitudes),
                units=operand.units,
            )
        case _:
            assert_never(unary_operator)


@functools.singledispatch
def negate(instruction) -> TimedInstruction[np.float64]:
    raise NotImplementedError(f"Negation of {type(instruction)} is not supported.")


@negate.register(Pattern)
def negate_pattern(pattern: Pattern[np.float64]) -> Pattern[np.float64]:
    return Pattern.create_without_copy(-pattern.array)


@negate.register(Concatenated)
def negate_concatenated(
    concatenated: Concatenated[np.float64],
) -> Concatenated[np.float64]:
    return Concatenated(*(negate(instr) for instr in concatenated.instructions))


@negate.register(Repeated)
def negate_repeated(repeated: Repeated[np.float64]) -> Repeated[np.float64]:
    return Repeated(repeated.repetitions, negate(repeated.instruction))


@negate.register(Ramp)
def negate_ramp(ramp: Ramp[np.float64]) -> TimedInstruction[np.float64]:
    return create_ramp(-ramp.start, -ramp.stop, len(ramp))


def multiply(
    a: TimedInstruction[np.float64], b: TimedInstruction[np.float64]
) -> TimedInstruction[np.float64]:
    merged = merge_instructions(left=a, right=b)
    return _multiply(merged)


@functools.singledispatch
def _multiply(instruction) -> TimedInstruction[np.float64]:
    raise NotImplementedError(
        f"Multiplication of {type(instruction)} is not supported."
    )


@_multiply.register(Pattern)
def _multiply_pattern(pattern: Pattern[np.void]) -> Pattern[np.float64]:
    left_array = pattern["left"].array
    right_array = pattern["right"].array
    return Pattern.create_without_copy(left_array * right_array)


@_multiply.register(Concatenated)
def _multiply_concatenated(
    concatenated: Concatenated[np.void],
) -> TimedInstruction[np.float64]:
    return concatenate(*(_multiply(instr) for instr in concatenated.instructions))


@_multiply.register(Repeated)
def _multiply_repeated(repeated: Repeated[np.void]) -> TimedInstruction[np.float64]:
    return repeated.repetitions * _multiply(repeated.instruction)


@_multiply.register(Ramp)
def _multiply_ramp(ramp: Ramp[np.void]) -> TimedInstruction[np.float64]:
    left = ramp["left"]
    right = ramp["right"]

    a2 = left.slope * right.slope
    if a2 == 0:
        a1 = left.slope * right.intercept + right.slope * left.intercept
        a0 = left.intercept * right.intercept
        stop = a0 + a1 * len(ramp)
        return create_ramp(a0, stop, len(ramp))
    else:
        return _multiply(ramp.to_pattern())


def divide(
    a: TimedInstruction[np.float64], b: TimedInstruction[np.float64]
) -> TimedInstruction[np.float64]:
    merged = merge_instructions(left=a, right=b)
    return _divide(merged)


@functools.singledispatch
def _divide(instruction) -> TimedInstruction[np.float64]:
    raise NotImplementedError(f"Division of {type(instruction)} is not supported.")


@_divide.register(Pattern)
def _divide_pattern(pattern: Pattern[np.void]) -> Pattern[np.float64]:
    left_array = pattern["left"].array
    right_array = pattern["right"].array
    if np.any(right_array == 0):
        raise ZeroDivisionError("Division by zero in pattern division.")
    result = left_array / right_array
    return Pattern.create_without_copy(result)


@_divide.register(Concatenated)
def _divide_concatenated(
    concatenated: Concatenated[np.void],
) -> TimedInstruction[np.float64]:
    return concatenate(*(_divide(instr) for instr in concatenated.instructions))


@_divide.register(Repeated)
def _divide_repeated(repeated: Repeated[np.void]) -> TimedInstruction[np.float64]:
    return repeated.repetitions * _divide(repeated.instruction)


@_divide.register(Ramp)
def _divide_ramp(ramp: Ramp[np.void]) -> TimedInstruction[np.float64]:
    left = ramp["left"]
    right = ramp["right"]

    if right.slope == 0:
        if right.intercept == 0:
            raise ZeroDivisionError("Division by zero in ramp division.")
        a1 = left.slope / right.intercept
        a0 = left.intercept / right.intercept
        stop = a0 + a1 * len(ramp)
        return create_ramp(a0, stop, len(ramp))
    else:
        return _divide(ramp.to_pattern())


def add(
    a: TimedInstruction[np.float64], b: TimedInstruction[np.float64]
) -> TimedInstruction[np.float64]:
    merged = merge_instructions(left=a, right=b)
    return _add(merged)


@functools.singledispatch
def _add(instruction) -> TimedInstruction[np.float64]:
    raise NotImplementedError(f"Addition of {type(instruction)} is not supported.")


@_add.register(Pattern)
def _add_pattern(pattern: Pattern[np.void]) -> Pattern[np.float64]:
    left_array = pattern["left"].array
    right_array = pattern["right"].array
    return Pattern.create_without_copy(left_array + right_array)


@_add.register(Concatenated)
def _add_concatenated(
    concatenated: Concatenated[np.void],
) -> TimedInstruction[np.float64]:
    return concatenate(*(_add(instr) for instr in concatenated.instructions))


@_add.register(Repeated)
def _add_repeated(repeated: Repeated[np.void]) -> TimedInstruction[np.float64]:
    return repeated.repetitions * _add(repeated.instruction)


@_add.register(Ramp)
def _add_ramp(ramp: Ramp[np.void]) -> TimedInstruction[np.float64]:
    left = ramp["left"]
    right = ramp["right"]

    a1 = left.slope + right.slope
    a0 = left.intercept + right.intercept
    stop = a0 + a1 * len(ramp)
    return create_ramp(a0, stop, len(ramp))



"""Contains the copy function to transfer data between two sessions."""

from typing import assert_never, assert_type

import tqdm

from caqtus.utils.result import (
    Success,
    Failure,
    is_failure,
    is_failure_type,
    unwrap,
)
from .._exception_summary import TracebackSummary
from .._experiment_session import ExperimentSession
from .._path import PureSequencePath
from .._path_hierarchy import PathNotFoundError, PathIsRootError, PathHasChildrenError
from .._sequence_collection import InvalidStateTransitionError
from .._sequence_collection import (
    PathIsNotSequenceError,
    PathIsSequenceError,
    SequenceStateError,
    SequenceNotCrashedError,
)
from .._session_maker import ExperimentSessionMaker
from .._shot_id import ShotId
from .._state import State


def copy_path(
    path: PureSequencePath,
    source: ExperimentSessionMaker,
    destination: ExperimentSessionMaker,
) -> None:
    """Copy the path and its descendants from one session to another.

    Args:
        path: The path to copy.
        source: Used to connect to the storage when the data to be copied is located.
            The path must exist in the source session.
        destination: Used to connect to the storage where the data will be copied.
            The path must not exist in the destination session.
    """

    with source() as source_session, destination() as destination_session:
        unwrap(_copy_path(path, source_session, destination_session))


def _copy_path(
    path: PureSequencePath,
    source_session: ExperimentSession,
    destination_session: ExperimentSession,
) -> (
    Success[None]
    | Failure[PathIsSequenceError]
    | Failure[PathNotFoundError]
    | Failure[SequenceStateError]
    | Failure[PathHasChildrenError]
):
    """Copy the path and its descendants from one session to another."""

    result = copy_paths_only(path, source_session, destination_session)
    if is_failure(result):
        return result
    sequences_to_copy = result.content()
    progress_bar = tqdm.tqdm(sequences_to_copy)
    for sequence in progress_bar:
        progress_bar.set_description(f"Copying {sequence}")
        copy_sequence_result = copy_sequence(
            sequence, source_session, destination_session
        )
        if is_failure(copy_sequence_result):
            return copy_sequence_result
    return Success(None)


def copy_paths_only(
    path: PureSequencePath,
    source_session: ExperimentSession,
    destination_session: ExperimentSession,
) -> (
    Success[list[PureSequencePath]]
    | Failure[PathIsSequenceError]
    | Failure[PathNotFoundError]
):
    """Copy the path and its descendants, ignoring sequences.

    Returns a list of sequences encountered inside the path.
    """

    path_creation_result = destination_session.paths.create_path(path)
    if is_failure(path_creation_result):
        return path_creation_result
    created_paths = path_creation_result.content()
    for created_path in created_paths:
        creation_date_result = source_session.paths.get_path_creation_date(created_path)
        assert not is_failure_type(creation_date_result, PathIsRootError)
        if is_failure(creation_date_result):
            return creation_date_result
        creation_date = creation_date_result.content()
        destination_session.paths.update_creation_date(created_path, creation_date)
    children_result = source_session.paths.get_children(path)
    if is_failure_type(children_result, PathNotFoundError):
        return children_result
    elif is_failure_type(children_result, PathIsSequenceError):
        return Success([path])
    else:
        children = children_result.content()
        result = []
        for child in children:
            copy_child_result = copy_paths_only(
                child, source_session, destination_session
            )
            if is_failure(copy_child_result):
                return copy_child_result
            result.extend(copy_child_result.content())
        return Success(result)


def copy_sequence(
    path: PureSequencePath,
    source_session: ExperimentSession,
    destination_session: ExperimentSession,
) -> Success[None] | Failure[SequenceStateError] | Failure[PathHasChildrenError]:
    sequence_stats_result = source_session.sequences.get_stats(path)
    assert not is_failure_type(sequence_stats_result, PathNotFoundError)
    assert not is_failure_type(sequence_stats_result, PathIsNotSequenceError)
    stats = sequence_stats_result.value
    state = stats.state
    if state in {State.RUNNING, State.PREPARING}:
        return Failure(SequenceStateError("Can't copy running sequence"))
    iterations = source_session.sequences.get_iteration_configuration(path)
    time_lanes = source_session.sequences.get_time_lanes(path)
    assert not is_failure_type(time_lanes, (PathNotFoundError, PathIsNotSequenceError))

    creation_result = destination_session.sequences.create(path, iterations, time_lanes)
    assert not is_failure_type(creation_result, PathIsSequenceError)
    if is_failure(creation_result):
        return creation_result

    if state == State.DRAFT:
        return Success(None)
    device_configs = source_session.sequences.get_device_configurations(path)
    global_parameters = unwrap(source_session.sequences.get_global_parameters(path))
    preparing_result = destination_session.sequences.set_preparing(
        path, device_configs, global_parameters
    )
    assert not is_failure_type(preparing_result, PathNotFoundError)
    assert not is_failure_type(preparing_result, PathIsNotSequenceError)
    assert not is_failure_type(preparing_result, InvalidStateTransitionError)
    assert_type(preparing_result, Success[None])

    assert stats.start_time is not None
    running_result = destination_session.sequences.set_running(
        path, start_time=stats.start_time
    )
    assert not is_failure_type(running_result, PathNotFoundError)
    assert not is_failure_type(running_result, PathIsNotSequenceError)

    for shot_index in tqdm.trange(stats.number_completed_shots):
        shot_parameters = source_session.sequences.get_shot_parameters(path, shot_index)
        shot_data = source_session.sequences.get_all_shot_data(path, shot_index)
        shot_start_time = source_session.sequences.get_shot_start_time(path, shot_index)
        shot_stop_time = source_session.sequences.get_shot_end_time(path, shot_index)
        unwrap(
            destination_session.sequences.create_shot(
                ShotId(path, shot_index),
                shot_parameters,
                shot_data,
                shot_start_time,
                shot_stop_time,
            )
        )

    if state == State.FINISHED:
        assert stats.stop_time is not None
        destination_session.sequences.set_finished(path, stop_time=stats.stop_time)
    elif state == State.INTERRUPTED:
        assert stats.stop_time is not None
        destination_session.sequences.set_interrupted(path, stop_time=stats.stop_time)
    elif state == State.CRASHED:
        exception_result = source_session.sequences.get_exception(path)
        assert not is_failure_type(exception_result, PathNotFoundError)
        assert not is_failure_type(exception_result, PathIsNotSequenceError)
        assert not is_failure_type(exception_result, SequenceNotCrashedError)
        exception = exception_result.value
        if exception is None:
            exception = TracebackSummary.from_exception(RuntimeError("Unknown error"))
        assert stats.stop_time is not None
        crashed_result = destination_session.sequences.set_crashed(
            path, exception, stop_time=stats.stop_time
        )
        assert not is_failure_type(crashed_result, PathNotFoundError)
        assert not is_failure_type(crashed_result, PathIsNotSequenceError)
        assert not is_failure_type(crashed_result, InvalidStateTransitionError)
        assert_type(crashed_result, Success[None])
    else:
        assert_never(state)  # type: ignore[reportArgumentType]

    return Success(None)



# Ignore some lint rules for this file as PySide6 models have a lot of camelCase
# methods.
# ruff: noqa: N802
from __future__ import annotations

import copy
from typing import Optional

import attrs
from PySide6.QtCore import (
    QAbstractListModel,
    QObject,
    QModelIndex,
    QPersistentModelIndex,
    Qt,
)
from PySide6.QtGui import QFont, QUndoCommand, QUndoStack

from caqtus.types.expression import Expression

_DEFAULT_INDEX: QModelIndex | QPersistentModelIndex = QModelIndex()


class TimeStepNameModel(QAbstractListModel):
    def __init__(self, undo_stak: QUndoStack, parent: Optional[QObject] = None):
        super().__init__(parent)
        self._names: list[str] = []
        self._undo_stack = undo_stak

    def set_names(self, names: list[str]):
        self.beginResetModel()
        self._names = copy.deepcopy(names)
        self.endResetModel()

    def get_names(self) -> list[str]:
        """Return a copy of the names displayed in the model."""

        return copy.deepcopy(self._names)

    def rowCount(
        self, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> int:
        return len(self._names)

    def data(self, index, role: int = Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
            return self._names[index.row()]
        if role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter

    def setData(self, index, value, role: int = Qt.ItemDataRole.EditRole) -> bool:
        if not index.isValid():
            return False
        if role == Qt.ItemDataRole.EditRole:
            if not isinstance(value, str):
                raise TypeError(f"Expected str, got {type(value)}")
            self._undo_stack.push(
                self.SetDataCommand(self, index.row(), self._names[index.row()], value)
            )
            return True
        return False

    @attrs.define(slots=False)
    class SetDataCommand(QUndoCommand):
        model: TimeStepNameModel
        row: int
        previous_value: str
        new_value: str

        def __attrs_post_init__(self):
            super().__init__(
                f"change name of Step 1 from <{self.previous_value}> to "
                f"<{self.new_value}>"
            )

        def redo(self) -> None:
            self.model._names[self.row] = self.new_value
            self.model.dataChanged.emit(
                self.model.index(self.row), self.model.index(self.row)
            )

        def undo(self) -> None:
            self.model._names[self.row] = self.previous_value
            self.model.dataChanged.emit(
                self.model.index(self.row), self.model.index(self.row)
            )

    def flags(self, index) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsEditable

    def insertRow(
        self, row, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> bool:
        if not (0 <= row <= self.rowCount()):
            return False
        self._undo_stack.push(self._InsertStepCommand(self, row))
        return True

    @attrs.define(slots=False)
    class _InsertStepCommand(QUndoCommand):
        model: TimeStepNameModel
        step: int

        def __attrs_post_init__(self):
            super().__init__(f"insert step {self.step}")

        def redo(self) -> None:
            self.model._insert_row_without_undo(self.step, f"Step {self.step}")

        def undo(self) -> None:
            self.model._remove_row_without_undo(self.step)

    def _insert_row_without_undo(self, row, value: str) -> None:
        assert 0 <= row <= self.rowCount()
        self.beginInsertRows(_DEFAULT_INDEX, row, row)
        self._names.insert(row, value)
        self.endInsertRows()

    def removeRow(
        self, row, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> bool:
        if not (0 <= row < self.rowCount()):
            return False
        self._undo_stack.push(self._RemoveStepCommand(self, row, self._names[row]))
        return True

    @attrs.define(slots=False)
    class _RemoveStepCommand(QUndoCommand):
        model: TimeStepNameModel
        step: int
        value: str

        def __attrs_post_init__(self):
            super().__init__(f"remove step {self.step}")

        def redo(self) -> None:
            self.model._remove_row_without_undo(self.step)

        def undo(self) -> None:
            self.model._insert_row_without_undo(self.step, self.value)

    def _remove_row_without_undo(self, row) -> None:
        assert 0 <= row < self.rowCount()
        self.beginRemoveRows(_DEFAULT_INDEX, row, row)
        del self._names[row]
        self.endRemoveRows()

    def headerData(self, section, orientation, role: int = Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return "Step names"
            elif orientation == Qt.Orientation.Vertical:
                return section
        elif role == Qt.ItemDataRole.FontRole:
            font = QFont()
            font.setBold(True)
            return font


class TimeStepDurationModel(QAbstractListModel):
    def __init__(self, undo_stack: QUndoStack, parent: Optional[QObject] = None):
        super().__init__(parent)
        self._durations: list[Expression] = []
        self._undo_stack = undo_stack

    def set_durations(self, durations: list[Expression]):
        self.beginResetModel()
        self._durations = copy.deepcopy(durations)
        self.endResetModel()

    def get_duration(self) -> list[Expression]:
        return copy.deepcopy(self._durations)

    def rowCount(
        self, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> int:
        return len(self._durations)

    def data(self, index, role: int = Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
            return self._durations[index.row()].body
        if role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter

    def setData(self, index, value, role: int = Qt.ItemDataRole.EditRole) -> bool:
        if not index.isValid():
            return False
        if role == Qt.ItemDataRole.EditRole:
            if not isinstance(value, str):
                raise TypeError(f"Expected str, got {type(value)}")
            self._undo_stack.push(
                self.SetDataCommand(
                    self, index.row(), self._durations[index.row()], Expression(value)
                )
            )
            return True
        return False

    @attrs.define(slots=False)
    class SetDataCommand(QUndoCommand):
        model: TimeStepDurationModel
        row: int
        previous_value: Expression
        new_value: Expression

        def __attrs_post_init__(self):
            super().__init__(
                f"change duration of Step 1 from <{self.previous_value}> to "
                f"<{self.new_value}>"
            )

        def redo(self) -> None:
            self.model._durations[self.row] = self.new_value
            self.model.dataChanged.emit(
                self.model.index(self.row), self.model.index(self.row)
            )

        def undo(self) -> None:
            self.model._durations[self.row] = self.previous_value
            self.model.dataChanged.emit(
                self.model.index(self.row), self.model.index(self.row)
            )

    def flags(self, index) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsEditable

    def insertRow(
        self, row, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> bool:
        if not (0 <= row <= self.rowCount()):
            return False
        self._undo_stack.push(self._InsertStepCommand(self, row))
        return True

    @attrs.define(slots=False)
    class _InsertStepCommand(QUndoCommand):
        model: TimeStepDurationModel
        step: int

        def __attrs_post_init__(self):
            super().__init__(f"insert step {self.step}")

        def redo(self) -> None:
            self.model._insert_row_without_undo(self.step, Expression("..."))

        def undo(self) -> None:
            self.model._remove_row_without_undo(self.step)

    def _insert_row_without_undo(self, row, value: Expression) -> None:
        assert 0 <= row <= self.rowCount()
        self.beginInsertRows(_DEFAULT_INDEX, row, row)
        self._durations.insert(row, value)
        self.endInsertRows()

    def removeRow(
        self, row, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> bool:
        if not (0 <= row < self.rowCount()):
            return False
        self._undo_stack.push(self._RemoveStepCommand(self, row, self._durations[row]))
        return True

    @attrs.define(slots=False)
    class _RemoveStepCommand(QUndoCommand):
        model: TimeStepDurationModel
        step: int
        value: Expression

        def __attrs_post_init__(self):
            super().__init__(f"remove step {self.step}")

        def redo(self) -> None:
            self.model._remove_row_without_undo(self.step)

        def undo(self) -> None:
            self.model._insert_row_without_undo(self.step, self.value)

    def _remove_row_without_undo(self, row) -> None:
        assert 0 <= row < self.rowCount()
        self.beginRemoveRows(_DEFAULT_INDEX, row, row)
        del self._durations[row]
        self.endRemoveRows()

    def headerData(self, section, orientation, role: int = Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return "Step durations"
            elif orientation == Qt.Orientation.Vertical:
                return section
        elif role == Qt.ItemDataRole.FontRole:
            font = QFont()
            font.setBold(True)
            return font



from .sequence_iteration_editor import SequenceIterationEditor, IterationEditorCreator
from .steps_iteration_editor import StepsIterationEditor

__all__ = [
    "SequenceIterationEditor",
    "IterationEditorCreator",
    "StepsIterationEditor",
]



import abc
from typing import runtime_checkable, Protocol, Self, ParamSpec

InitParams = ParamSpec("InitParams")


@runtime_checkable
class Device(Protocol[InitParams]):
    """Wraps a low-level instrument that can be controlled during an experiment.

    Subclasses of this class should at least implement the :meth:`__init__`,
    :meth:`__enter__`, and :meth:`__exit__` methods.
    In addition to these methods, any other methods that are specific to the device
    should be implemented.
    """

    def __init__(self, *args: InitParams.args, **kwargs: InitParams.kwargs) -> None:
        """Device constructor.

        No communication to an instrument or initialization should be done in the
        constructor.
        Instead, use the :meth:`__enter__` method to acquire the necessary resources.
        """

        pass

    @abc.abstractmethod
    def __enter__(self) -> Self:
        """Initialize the device.

        Used to establish communication to the device and allocate the necessary
        resources.

        Warnings:
            If you need to acquire multiple resources in the :meth:`__enter__` method,
            you need to ensure that the first resources are correctly released if an
            error occurs while acquiring the subsequent resources.

            If this is the case, you can use this pattern: :ref:`clean-up-enter`.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Shutdown the device.

        Used to terminate communication to the device and free the associated resources.
        """

        raise NotImplementedError



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import attrs

from caqtus.shot_compilation import ShotContext
from caqtus.types.expression import Expression
from caqtus.types.recoverable_exceptions import InvalidValueError
from caqtus.types.units import (
    is_scalar_quantity,
    DimensionalityError,
    InvalidDimensionalityError,
    NANOSECOND,
)
from caqtus.types.variable_name import DottedVariableName
from ..channel_output import ChannelOutput
from ...timing import TimeStep


@attrs.define
class Advance(ChannelOutput):
    input_: ChannelOutput = attrs.field(
        validator=attrs.validators.instance_of(ChannelOutput),
        on_setattr=attrs.setters.validate,
    )
    advance: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )

    def __str__(self):
        return f"{self.input_} << {self.advance}"

    def evaluate(
        self,
        required_time_step: TimeStep,
        prepend: int,
        append: int,
        shot_context: ShotContext,
    ):
        evaluated_advance = _evaluate_in_nanoseconds(
            self.advance, shot_context.get_parameters()
        )
        number_ticks_to_advance = round(evaluated_advance / float(required_time_step))
        if number_ticks_to_advance < 0:
            raise ValueError(
                f"Cannot advance by a negative number of time steps "
                f"({number_ticks_to_advance})"
            )
        if number_ticks_to_advance > prepend:
            raise ValueError(
                f"Cannot advance by {number_ticks_to_advance} time steps when only "
                f"{prepend} are available"
            )
        return self.input_.evaluate(
            required_time_step,
            prepend - number_ticks_to_advance,
            append + number_ticks_to_advance,
            shot_context,
        )

    def evaluate_max_advance_and_delay(
        self,
        time_step: TimeStep,
        variables: Mapping[DottedVariableName, Any],
    ) -> tuple[int, int]:
        advance = _evaluate_in_nanoseconds(self.advance, variables)
        if advance < 0:
            raise InvalidValueError("Advance must be a positive number.")
        advance_ticks = round(advance / float(time_step))
        input_advance, input_delay = self.input_.evaluate_max_advance_and_delay(
            time_step, variables
        )
        return advance_ticks + input_advance, input_delay


@attrs.define
class Delay(ChannelOutput):
    input_: ChannelOutput = attrs.field(
        validator=attrs.validators.instance_of(ChannelOutput),
        on_setattr=attrs.setters.validate,
    )
    delay: Expression = attrs.field(
        validator=attrs.validators.instance_of(Expression),
        on_setattr=attrs.setters.validate,
    )

    def __str__(self):
        return f"{self.delay} >> {self.input_}"

    def evaluate(
        self,
        required_time_step: TimeStep,
        prepend: int,
        append: int,
        shot_context: ShotContext,
    ):
        evaluated_delay = _evaluate_in_nanoseconds(
            self.delay, shot_context.get_parameters()
        )
        number_ticks_to_delay = round(evaluated_delay / float(required_time_step))
        if number_ticks_to_delay < 0:
            raise ValueError(
                f"Cannot delay by a negative number of time steps "
                f"({number_ticks_to_delay})"
            )
        if number_ticks_to_delay > append:
            raise ValueError(
                f"Cannot delay by {number_ticks_to_delay} time steps when only "
                f"{append} are available"
            )
        return self.input_.evaluate(
            required_time_step,
            prepend + number_ticks_to_delay,
            append - number_ticks_to_delay,
            shot_context,
        )

    def evaluate_max_advance_and_delay(
        self,
        time_step: TimeStep,
        variables: Mapping[DottedVariableName, Any],
    ) -> tuple[int, int]:
        delay = _evaluate_in_nanoseconds(self.delay, variables)
        if delay < 0:
            raise ValueError("Delay must be a positive number.")
        delay_ticks = round(delay / float(time_step))
        input_advance, input_delay = self.input_.evaluate_max_advance_and_delay(
            time_step, variables
        )
        return input_advance, delay_ticks + input_delay


def _evaluate_in_nanoseconds(
    expression: Expression, variables: Mapping[DottedVariableName, Any]
) -> float:
    evaluated = expression.evaluate(variables)
    if not is_scalar_quantity(evaluated):
        raise InvalidValueError("Advance must be a scalar quantity.")
    try:
        evaluated_advance = evaluated.to_unit(NANOSECOND).magnitude
    except DimensionalityError as e:
        raise InvalidDimensionalityError(
            f"Advance must be expressed in seconds, not {evaluated.units}"
        ) from e
    return evaluated_advance



"""Contains utilities to generate complex values for the output of a device.

It is convenient to transform input values given by the user into output values that
a device should generate.
This module contains classes that can be used to construct complex tree structures that
represent user defined transformations.
"""

from ._converter import converter, structure_evaluable_output
from ._output_mapping import LinearInterpolation
from ._transformation import (
    Transformation,
    evaluate,
    EvaluableOutput,
    evaluable_output_validator,
)


__all__ = [
    "Transformation",
    "LinearInterpolation",
    "evaluate",
    "EvaluableOutput",
    "converter",
    "evaluable_output_validator",
    "structure_evaluable_output",
]



import functools
from typing import assert_never

import caqtus_parsing.nodes as nodes


@functools.lru_cache
def is_time_dependent(expression: nodes.Expression) -> bool:
    match expression:
        case int() | float() | nodes.Quantity():
            return False
        case nodes.Variable(name=name):
            return name == "t"
        case (
            nodes.Add()
            | nodes.Subtract()
            | nodes.Multiply()
            | nodes.Divide()
            | nodes.Power() as binary_operator
        ):
            return is_time_dependent(binary_operator.left) or is_time_dependent(
                binary_operator.right
            )
        case nodes.Plus() | nodes.Minus() as unary_operator:
            return is_time_dependent(unary_operator.operand)
        case nodes.Call():
            return any(is_time_dependent(arg) for arg in expression.args)
        case _:
            assert_never(expression)



import decimal

import cattrs.strategies
from cattrs.gen import make_dict_structure_fn, override

from caqtus.types.expression import Expression
from caqtus.utils import serialization
from .channel_commands import (
    CalibratedAnalogMapping,
    ChannelOutput,
    LaneValues,
    Constant,
    DeviceTrigger,
)
from .channel_commands.timing import Advance, Delay, BroadenLeft
from .timing import TimeStep
from .trigger import TriggerEdge, Trigger

converter = serialization.copy_converter()
"""A converter than can serialize and deserialize sequencer configuration."""


def _structure_time_step(value, _) -> TimeStep:
    return TimeStep(decimal.Decimal(value))


converter.register_structure_hook(TimeStep, _structure_time_step)


def _unstructure_time_step(value) -> str:
    return str(value)


converter.register_unstructure_hook(TimeStep, _unstructure_time_step)


converter.register_unstructure_hook(TriggerEdge, lambda edge: edge.value)
cattrs.strategies.configure_tagged_union(Trigger, converter, tag_name="trigger type")


# This is a legacy workaround for structuring ChannelOutput subclasses.
# This is because at the beginning of the project, the type field was not serialized
# but instead the type was inferred from the structure of the data.
# For new data, we can use the type field to determine the type of the
# ChannelOutput subclass, but for old data, we need to infer the type from the
# fields present in the data.
def structure_channel_output(data, _):
    if "type" in data:
        return converter.structure(data, ChannelOutput)
    elif "lane" in data:
        return converter.structure(data, LaneValues)
    elif data.keys() == {"value"}:
        return converter.structure(data, Constant)
    else:
        raise ValueError(f"Cannot structure {data} as a ChannelOutput")


structure_hook = cattrs.gen.make_dict_structure_fn(
    CalibratedAnalogMapping,
    converter,
    input_=cattrs.override(struct_hook=structure_channel_output),
)

converter.register_structure_hook(CalibratedAnalogMapping, structure_hook)

advance_structure_hook = cattrs.gen.make_dict_structure_fn(
    Advance,
    converter,
    input_=cattrs.override(struct_hook=structure_channel_output),
)

converter.register_structure_hook(Advance, advance_structure_hook)

delay_structure_hook = cattrs.gen.make_dict_structure_fn(
    Delay,
    converter,
    input_=cattrs.override(struct_hook=structure_channel_output),
)

converter.register_structure_hook(Delay, delay_structure_hook)

broaden_left_structure_hook = cattrs.gen.make_dict_structure_fn(
    BroadenLeft,
    converter,
    input_=cattrs.override(struct_hook=structure_channel_output),
)

converter.register_structure_hook(BroadenLeft, broaden_left_structure_hook)


def structure_default(data, _):
    # We need this custom structure hook, because in the past the default value of a
    # DeviceTrigger was a Constant and not any ChannelOutput.
    # In that case, the type of the default value was not serialized, so we need to
    # deal with this special case.
    if data is None:
        return None
    if "type" in data:
        return converter.structure(data, ChannelOutput)
    else:
        return converter.structure(data, Constant)


structure_device_trigger = make_dict_structure_fn(
    DeviceTrigger,
    converter,
    default=override(struct_hook=structure_default),
)
converter.register_structure_hook(DeviceTrigger, structure_device_trigger)


def structure_lane_default(default_data, _):
    # We need this custom structure hook, because in the past the default value of a
    # LaneValues was a Constant and not any ChannelOutput.
    # In that case, the type of the default value was not serialized, so we need to
    # deal with this special case.
    if default_data is None:
        return None
    elif isinstance(default_data, str):
        default_expression = converter.structure(default_data, Expression)
        return Constant(value=default_expression)
    elif "type" in default_data:
        return converter.structure(default_data, ChannelOutput)
    else:
        return converter.structure(default_data, Constant)


structure_lane_values = make_dict_structure_fn(
    LaneValues,
    converter,
    default=override(struct_hook=structure_lane_default),
)
converter.register_structure_hook(LaneValues, structure_lane_values)


def union_strategy(union, converter):
    return cattrs.strategies.configure_tagged_union(union, converter, tag_name="type")


cattrs.strategies.include_subclasses(
    ChannelOutput, converter, union_strategy=union_strategy
)



from __future__ import annotations

import functools
from typing import assert_never, Optional

import anyio
import anyio.abc
import anyio.lowlevel
from PySide6 import QtCore
from PySide6.QtCore import (
    QModelIndex,
)
from PySide6.QtWidgets import (
    QMenu,
    QMessageBox,
    QInputDialog,
    QLineEdit,
    QApplication,
    QAbstractItemView,
)

from caqtus.gui._common.sequence_hierarchy import (
    AsyncPathHierarchyView,
)
from caqtus.gui.qtutil import temporary_widget
from caqtus.session import (
    ExperimentSessionMaker,
    PureSequencePath,
    PathIsRootError,
    PathNotFoundError,
)
from caqtus.session import (
    PathIsSequenceError,
    PathHasChildrenError,
    State,
)
from caqtus.types.expression import Expression
from caqtus.types.iteration import (
    StepsConfiguration,
    ArangeLoop,
    ExecuteShot,
)
from caqtus.types.timelane import TimeLanes
from caqtus.types.variable_name import DottedVariableName
from caqtus.utils.result import Failure, is_success, is_failure_type, unwrap
from caqtus.utils.result._result import is_failure
from ._icons import get_icon
from .._common.waiting_widget import blocking_call


class EditablePathHierarchyView(AsyncPathHierarchyView):
    sequence_start_requested = QtCore.Signal(PureSequencePath)
    sequence_interrupt_requested = QtCore.Signal(PureSequencePath)

    def __init__(
        self,
        session_maker: ExperimentSessionMaker,
        parent=None,
    ):
        super().__init__(session_maker, parent)
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_context_menu)  # type: ignore
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.setDragEnabled(True)
        self.viewport().setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
        self._task_group: Optional[anyio.abc.TaskGroup] = None

    async def run_async(self) -> None:
        async with anyio.create_task_group() as self._task_group:
            await super().run_async()

    def show_context_menu(self, pos):
        proxy_index = self.indexAt(pos)
        index = self._proxy_model.mapToSource(proxy_index)

        path = self._model.get_path(index)

        with temporary_widget(QMenu(self)) as menu:
            color = self.palette().text().color()

            with self.session_maker() as session:
                is_sequence = unwrap(session.sequences.is_sequence(path))
                if is_sequence:
                    state = unwrap(session.sequences.get_state(path))
                else:
                    state = None

            if not is_sequence:
                new_menu = menu.addMenu("New...")

                create_folder_action = new_menu.addAction("folder")
                create_folder_action.triggered.connect(
                    functools.partial(self.create_new_folder, path)
                )

                create_sequence_action = new_menu.addAction("sequence")
                create_sequence_action.triggered.connect(
                    functools.partial(self.create_new_sequence, index)
                )
            if is_sequence:
                play_icon = get_icon("start", color)
                start_action = menu.addAction(play_icon, "Start")
                if state == State.DRAFT:
                    start_action.setEnabled(True)
                else:
                    start_action.setEnabled(False)
                start_action.triggered.connect(
                    lambda: self.sequence_start_requested.emit(path)
                )

                stop_icon = get_icon("stop", color)
                stop_action = menu.addAction(stop_icon, "Interrupt")
                if state == State.RUNNING:
                    stop_action.setEnabled(True)
                else:
                    stop_action.setEnabled(False)
                stop_action.triggered.connect(
                    lambda: self.sequence_interrupt_requested.emit(path)
                )

                duplicate_icon = get_icon("duplicate", color)
                duplicate_action = menu.addAction(duplicate_icon, "Duplicate")
                duplicate_action.triggered.connect(
                    functools.partial(self.on_sequence_duplication_requested, index)
                )

                clear_icon = get_icon("clear", color)
                clear_action = menu.addAction(clear_icon, "Clear")
                clear_action.triggered.connect(
                    functools.partial(self.on_clear_sequence_requested, path)
                )
                if state not in {
                    State.FINISHED,
                    State.INTERRUPTED,
                    State.CRASHED,
                }:
                    clear_action.setEnabled(False)

            if not path.is_root():
                trash_icon = get_icon("delete", color)
                delete_action = menu.addAction(trash_icon, "Delete")
                if state not in {
                    State.DRAFT,
                    State.FINISHED,
                    State.INTERRUPTED,
                    State.CRASHED,
                    None,
                }:
                    delete_action.setEnabled(False)
                delete_action.triggered.connect(functools.partial(self.delete, index))
            if index.isValid():
                rename_action = menu.addAction(
                    get_icon("mdi6.form-textbox", color), "Rename"
                )
                rename_action.triggered.connect(
                    functools.partial(self.on_rename_requested, index)
                )

            menu.exec(self.mapToGlobal(pos))

    def on_rename_requested(self, index: QModelIndex) -> None:
        """Ask the user for a new name and rename the sequence."""

        app = QApplication.instance()
        if app is None:
            raise RuntimeError("No QApplication instance")

        path = self._model.get_path(index)

        assert path.name is not None

        text, ok = QInputDialog().getText(
            self,
            f"Rename {path}...",
            "New name:",
            QLineEdit.EchoMode.Normal,
            path.name,
        )
        if ok and text:
            if PureSequencePath.is_valid_name(text):
                result = self._model.rename(index, text)
                if isinstance(result, Failure):
                    QMessageBox.warning(  # type: ignore[reportCallIssue]
                        self,
                        app.applicationName(),
                        f"<p>Could not rename <i>{path.name}</i> to "
                        f"<i>{text}</i>:</p>"
                        f"<p>{result}</p>",
                    )
            else:
                QMessageBox.warning(  # type: ignore[reportCallIssue]
                    self,
                    app.applicationName(),
                    f"Name <i>{text}</i> is not valid.",
                )
                self.on_rename_requested(index)

    def on_clear_sequence_requested(self, path: PureSequencePath) -> None:
        """Clear the sequence at the given path.

        This will revert the sequence to the draft state, effectively clearing all
        the data in it.
        During the process, a waiting widget will be shown to the user to prevent them
        from interacting with the sequence while it is being cleared.
        """

        async def clear():
            async with self.session_maker.async_session() as session:
                await session.sequences.reset_to_draft(path)
                await anyio.lowlevel.checkpoint()

        if self._task_group is None:
            raise RuntimeError("Task group not initialized")

        self._task_group.start_soon(
            blocking_call,
            self,
            f"<p>Clearing <b><i>{path}</i></b></p><p>This may take a while...</p>",
            clear,
        )

    def on_sequence_duplication_requested(self, source: QModelIndex):
        """Ask the user for a new sequence name and duplicate the sequence."""

        path = self._model.get_path(source)

        assert path.name is not None

        text, ok = QInputDialog().getText(
            self,
            f"Duplicate {path}...",
            "New sequence name:",
            QLineEdit.EchoMode.Normal,
            path.name,
        )
        app = QApplication.instance()
        if app is None:
            raise RuntimeError("No QApplication instance")
        title = app.applicationName()
        if ok and text:
            with self.session_maker() as session:
                iterations = session.sequences.get_iteration_configuration(path)
                time_lanes = session.sequences.get_time_lanes(path)
                if is_failure(time_lanes):
                    return
            if text.startswith(PureSequencePath._separator()):
                if not PureSequencePath.is_valid_path(text):
                    QMessageBox.warning(  # type: ignore[reportCallIssue]
                        self,
                        title,
                        f"Path '{text}' is not valid for a sequence.",
                    )
                    return
                path = PureSequencePath(text)
                with self.session_maker() as session:
                    creation_result = session.sequences.create(
                        path, iterations, time_lanes
                    )
            else:
                if not PureSequencePath.is_valid_name(text):
                    QMessageBox.warning(  # type: ignore[reportCallIssue]
                        self,
                        title,
                        f"Name '{text}' is not valid for a sequence.",
                    )
                    return
                creation_result = self._model.create_new_sequence(
                    source.parent(), text, iterations, time_lanes
                )
            if is_success(creation_result):
                return
            elif is_failure_type(creation_result, PathIsSequenceError):
                QMessageBox.warning(  # type: ignore[reportCallIssue]
                    self,
                    title,
                    f"Target <i>{text}</i> already exists and is a sequence.",
                )
                return
            elif is_failure_type(creation_result, PathHasChildrenError):
                QMessageBox.warning(  # type: ignore[reportCallIssue]
                    self,
                    title,
                    f"Target <i>{text}</i> already exists and has children.",
                )
                return
            assert_never(creation_result)

    def create_new_folder(self, path: PureSequencePath):
        text, ok = QInputDialog().getText(
            self,
            f"New folder in {path}...",
            "Folder name:",
            QLineEdit.EchoMode.Normal,
            "new folder",
        )
        if ok and text:
            new_path = path / text
            with self.session_maker() as session:
                session.paths.create_path(new_path)

    def create_new_sequence(self, parent: QModelIndex):
        path = self._model.get_path(parent)
        text, ok = QInputDialog().getText(
            self,
            f"New sequence in {path}...",
            "Sequence name:",
            QLineEdit.EchoMode.Normal,
            "new sequence",
        )
        if ok and text:
            self._model.create_new_sequence(
                parent, text, DEFAULT_ITERATION_CONFIG, DEFAULT_TIME_LANES
            )

    def delete(self, index: QModelIndex):
        path = self._model.get_path(index)
        message = (
            f'You are about to delete "{path}".\n'
            "All data inside will be irremediably lost."
        )
        if self.exec_confirmation_message_box(message):
            result = self._model.remove_path(index)

            assert not is_failure_type(result, PathIsRootError)

            if is_success(result):
                return
            if is_failure_type(result, PathNotFoundError):
                return  # The path was already deleted
            if is_failure_type(result, PathIsSequenceError):
                app = QApplication.instance()
                if app is None:
                    raise RuntimeError("No QApplication instance")
                QMessageBox.warning(  # type: ignore[reportCallIssue]
                    self,
                    app.applicationName(),
                    f"The path '{path}' contains sequences and therefore "
                    f"cannot be deleted",
                )
                return
            assert_never(result)

    def exec_confirmation_message_box(self, message: str) -> bool:
        """Show a popup box to ask  a question."""

        with temporary_widget(QMessageBox(self)) as message_box:
            app = QApplication.instance()
            if app is None:
                raise RuntimeError("No QApplication instance")
            message_box.setWindowTitle(app.applicationName())
            message_box.setText(message)
            message_box.setInformativeText("Are you really sure you want to continue?")
            message_box.setStandardButtons(
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.Cancel
            )
            message_box.setDefaultButton(QMessageBox.StandardButton.Cancel)
            message_box.setIcon(QMessageBox.Icon.Warning)
            result = message_box.exec()
            if result == QMessageBox.StandardButton.Cancel:
                return False
            return True


DEFAULT_ITERATION_CONFIG = StepsConfiguration(
    steps=[
        ArangeLoop(
            variable=DottedVariableName("rep"),
            start=Expression("0"),
            stop=Expression("10"),
            step=Expression("1"),
            sub_steps=[ExecuteShot()],
        ),
    ]
)

DEFAULT_TIME_LANES = TimeLanes(
    step_names=["step 0"],
    step_durations=[Expression("...")],
    lanes={},
)



import datetime

import polars

from caqtus.session import Shot
from ._combinable_importers import CombinableLoader


class LoadShotTime(CombinableLoader):
    """Loads the time of a shot.

    When it is evaluated on a shot, it returns a polars dataframe with a single row and
    two columns: `start time` and `end time` with dtype `polars.Datetime` indicates
    when the shot started and ended.
    """

    def load(self, shot: Shot):
        return self._shot_time_to_dataframe(shot.get_start_time(), shot.get_end_time())

    @staticmethod
    def _shot_time_to_dataframe(
        start_time: datetime.datetime, stop_time: datetime.datetime
    ) -> polars.DataFrame:
        dataframe = polars.DataFrame(
            [
                polars.Series("start time", [start_time], dtype=polars.Datetime),
                polars.Series("end time", [stop_time], dtype=polars.Datetime),
            ]
        )
        return dataframe



from typing import Optional, Any, assert_never

from PySide6.QtCore import QObject, QModelIndex, Qt, QPersistentModelIndex
from PySide6.QtGui import QPalette

from caqtus.gui.condetrol._icons import get_icon
from caqtus.types.data import DataLabel
from caqtus.types.image import ImageLabel
from caqtus.types.timelane import CameraTimeLane, TakePicture
from .._time_lane_model import TimeLaneModel

_DEFAULT_INDEX = QModelIndex()


class CameraTimeLaneModel(TimeLaneModel[CameraTimeLane]):
    # ruff: noqa: N802
    def __init__(self, name: str, parent: Optional[QObject] = None):
        lane = CameraTimeLane([None])
        super().__init__(name, lane, parent)
        self._brush = None
        palette = QPalette()
        color = palette.text().color()
        self._icon = get_icon("camera", color=color)

    def data(self, index, role: int = Qt.ItemDataRole.DisplayRole):
        if not index.isValid():
            return None
        value = self.lane_value(index.row())
        if role == Qt.ItemDataRole.DisplayRole:
            if isinstance(value, TakePicture):
                return value.picture_name
            elif value is None:
                return None
            else:
                assert_never(value)
        elif role == Qt.ItemDataRole.EditRole:
            if isinstance(value, TakePicture):
                return value.picture_name
            elif value is None:
                return ""
            else:
                assert_never(value)
        elif role == Qt.ItemDataRole.ForegroundRole:
            return self._brush
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter
        elif role == Qt.ItemDataRole.DecorationRole:
            if isinstance(value, TakePicture):
                return self._icon
        else:
            return None

    def setData(self, index, value: Any, role: int = Qt.ItemDataRole.EditRole):
        if not index.isValid():
            return False
        if role == Qt.ItemDataRole.EditRole:
            assert isinstance(value, str)

            if value == "":
                new_value = None
            else:
                new_value = TakePicture(ImageLabel(DataLabel(value)))

            return self.set_lane_value(index.row(), new_value)

        return False

    def insertRow(
        self, row, parent: QModelIndex | QPersistentModelIndex = _DEFAULT_INDEX
    ) -> bool:
        return self.insert_lane_value(row, None)



from abc import ABCMeta
from typing import final


@final  # No subclassing of NoPublicConstructor itself.
class NoPublicConstructor(ABCMeta):
    """Metaclass that ensures a private constructor.

    If a class uses this metaclass like this::

        @final
        class SomeClass(metaclass=NoPublicConstructor):
            pass

    The metaclass will ensure that no instance can be initialized. This should always be
    used with @final.

    If you try to instantiate your class (SomeClass()), a TypeError will be thrown. Use
    _create() instead in the class's implementation.

    Raises
    ------
    - TypeError if an instance is created.
    """

    def __call__(cls, *args: object, **kwargs: object) -> None:
        raise TypeError(
            f"{cls.__module__}.{cls.__qualname__} has no public constructor"
        )

    def _create[T](cls: type[T], *args: object, **kwargs: object) -> T:
        return super().__call__(*args, **kwargs)  # type: ignore



from typing import NewType

from ._units import Unit, BaseUnit

Second = NewType("Second", BaseUnit)
SECOND = Second(Unit("s").to_base())
NANOSECOND = Unit("ns")

HERTZ = Unit("Hz")
MEGAHERTZ = Unit("MHz")

DECIBEL = Unit("dB")

VOLT = Unit("V")
AMPERE = Unit("A")



from __future__ import annotations

import abc
from collections.abc import Mapping, Callable
from typing import Any
from typing import Protocol

import anyio.to_process

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.shot_compilation import (
    SequenceContext,
    DeviceCompiler,
    DeviceNotUsedException,
)
from caqtus.shot_compilation.compilation_contexts import ShotContext
from caqtus.types._parameter_namespace import VariableNamespace
from caqtus.types.recoverable_exceptions import InvalidValueError
from caqtus.types.timelane import TimeLanes
from caqtus.utils._tblib import ensure_exception_pickling
from ._shot_primitives import ShotParameters
from ..device_manager_extension import DeviceManagerExtensionProtocol


class ShotCompilerProtocol(Protocol):
    @abc.abstractmethod
    def compile_initialization_parameters(
        self,
    ) -> Mapping[DeviceName, Mapping[str, Any]]:
        """Compile the initialization parameters for the devices.

        Returns:
            A mapping from device names to the initialization parameters for each
            device.

            The keys of the mapping are the device names that are in use for this
            sequence.
        """

        raise NotImplementedError

    @abc.abstractmethod
    async def compile_shot(
        self, shot_parameters: ShotParameters
    ) -> tuple[Mapping[DeviceName, Mapping[str, Any]], float]:
        raise NotImplementedError


type ShotCompilerFactory = Callable[
    [SequenceContext, DeviceManagerExtensionProtocol], ShotCompilerProtocol
]


class ShotCompiler(ShotCompilerProtocol):
    def __init__(
        self,
        shot_timelanes: TimeLanes,
        device_configurations: Mapping[DeviceName, DeviceConfiguration],
        device_compilers: Mapping[DeviceName, DeviceCompiler],
    ):
        self.shot_time_lanes = shot_timelanes
        self.device_configurations = device_configurations
        self._sequence_context = SequenceContext(
            device_configurations=device_configurations,  # pyright: ignore[reportCallIssue]
            time_lanes=shot_timelanes,  # pyright: ignore[reportCallIssue]
        )
        self.device_compilers = device_compilers

    def compile_initialization_parameters(
        self,
    ) -> Mapping[DeviceName, Mapping[str, Any]]:
        initialization_parameters = {}
        for device_name, compiler in self.device_compilers.items():
            initialization_parameters[device_name] = (
                compiler.compile_initialization_parameters()
            )
        return initialization_parameters

    async def compile_shot(
        self, shot_parameters: ShotParameters
    ) -> tuple[Mapping[DeviceName, Mapping[str, Any]], float]:
        # We add a deadline to shot compilation to not hang indefinitely in case of
        # a bug.
        with anyio.move_on_after(10):
            return await anyio.to_process.run_sync(
                self.compile_shot_sync, shot_parameters.parameters, cancellable=True
            )
        raise TimeoutError(
            "Shot compilation took too long to complete and has been terminated."
        )

    @ensure_exception_pickling
    def compile_shot_sync(
        self, shot_parameters: VariableNamespace
    ) -> tuple[Mapping[DeviceName, Mapping[str, Any]], float]:
        shot_context = ShotContext(
            sequence_context=self._sequence_context,  # pyright: ignore[reportCallIssue]
            variables=shot_parameters.dict(),  # pyright: ignore[reportCallIssue]
            device_compilers=self.device_compilers,  # pyright: ignore[reportCallIssue]
        )

        results = {}
        for device_name in self.device_compilers:
            results[device_name] = shot_context.get_shot_parameters(device_name)

        # noinspection PyProtectedMember
        if unused_lanes := shot_context._unused_lanes():
            raise InvalidValueError(
                "The following lanes where not used during the shot: "
                + ", ".join(unused_lanes)
            )

        return results, float(shot_context.get_shot_duration())


def create_shot_compiler(
    initial_sequence_context: SequenceContext,
    device_manager_extension: DeviceManagerExtensionProtocol,
) -> ShotCompiler:
    device_compilers = create_device_compilers(
        initial_sequence_context, device_manager_extension
    )
    in_use_configurations = {
        device_name: initial_sequence_context.get_device_configuration(device_name)
        for device_name in device_compilers
    }
    shot_compiler = _create_shot_compiler(
        time_lanes=initial_sequence_context._time_lanes,  # noqa
        device_configurations=in_use_configurations,
        device_compilers=device_compilers,
    )
    return shot_compiler


def create_device_compilers(
    sequence_context: SequenceContext,
    device_manager_extension: DeviceManagerExtensionProtocol,
) -> dict[DeviceName, DeviceCompiler]:
    device_compilers = {}
    for (
        device_name,
        device_configuration,
    ) in sequence_context.get_all_device_configurations().items():
        compiler_type = device_manager_extension.get_device_compiler_type(
            device_configuration
        )
        try:
            compiler = compiler_type(device_name, sequence_context)
        except DeviceNotUsedException:
            continue
        else:
            device_compilers[device_name] = compiler
    return device_compilers


def _create_shot_compiler(
    time_lanes: TimeLanes,
    device_configurations: Mapping[DeviceName, DeviceConfiguration],
    device_compilers: Mapping[DeviceName, DeviceCompiler],
) -> ShotCompiler:
    shot_compiler = ShotCompiler(
        time_lanes,
        device_configurations=device_configurations,
        device_compilers=device_compilers,
    )
    return shot_compiler



from typing import TypeVar, Callable

from attrs import AttrsInstance
from cattrs.gen import (
    AttributeOverride,  # pyright: ignore[reportPrivateImportUsage]
    make_dict_unstructure_fn,
    make_dict_structure_fn,
)

from .converters import converters

_T = TypeVar("_T", bound=AttrsInstance)


def customize(**kwargs: AttributeOverride) -> Callable[[type[_T]], type[_T]]:
    def decorator(cls: type[_T]) -> type[_T]:
        for converter in converters.values():
            unstructure_hook = make_dict_unstructure_fn(
                cls, converter, **kwargs  # pyright: ignore[reportArgumentType]
            )
            converter.register_unstructure_hook(cls, unstructure_hook)

            structure_hook = make_dict_structure_fn(
                cls, converter, **kwargs  # pyright: ignore[reportArgumentType]
            )
            converter.register_structure_hook(cls, structure_hook)

        return cls

    return decorator



from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import Protocol, TYPE_CHECKING

from caqtus.utils.result import Success, Failure
from ._path import PureSequencePath

if TYPE_CHECKING:
    from ._sequence_collection import PathIsSequenceError, SequenceRunningError


class PathHierarchy(Protocol):
    """A file-system-like object that exists inside a session.

    This object provides methods that can be used to create, delete and check for the
    existence of sequences.
    """

    @abstractmethod
    def does_path_exists(self, path: PureSequencePath) -> bool:
        """Check if the path exists in the session.

        Args:
            path: the path to check for existence.

        Returns:
            True if the path exists in the session. False otherwise.
        """

        raise NotImplementedError

    @abstractmethod
    def create_path(
        self, path: PureSequencePath
    ) -> Success[list[PureSequencePath]] | Failure[PathIsSequenceError]:
        """Create the path in the session and its ancestors if they do not exist.

        If is safe to call this method even if the path already exists, in which case
        it will return a Success with an empty list.

        Args:
            path: the path to create.

        Returns:
            * Success, with a list of the paths that were created if the path was
              created successfully. The list is ordered from parent to child.
            * Failure, with PathIsSequenceError if the path or one of its ancestors is a
              sequence. No path is created if any of the ancestors is a sequence.

              If a failure is returned, not paths are created.
        """

        raise NotImplementedError

    @abstractmethod
    def delete_path(
        self, path: PureSequencePath, delete_sequences: bool = False
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsSequenceError]
        | Failure[PathIsRootError]
    ):
        """Delete the path and all its descendants.

        Warnings:
            If delete_sequences is True, all sequences in the path will be deleted.

        Args:
            path: The path to delete. Descendants will be deleted recursively.
            delete_sequences: If False, raise an error if the path or one of its
            children is a sequence.

        Returns:
            Success, if the path was deleted successfully.

            Failure, with one of the following errors:

            * PathNotFoundError: If the path does not exist.
            * PathIsSequenceError: If the path or one of its children is a sequence and
                delete_sequence is False.
            * PathIsRootError: If the path is the root path.
        """

        raise NotImplementedError

    @abstractmethod
    def get_children(
        self, path: PureSequencePath
    ) -> (
        Success[set[PureSequencePath]]
        | Failure[PathNotFoundError]
        | Failure[PathIsSequenceError]
    ):
        """Get the children of the path."""

        raise NotImplementedError

    @abstractmethod
    def get_path_creation_date(
        self, path: PureSequencePath
    ) -> Success[datetime] | Failure[PathNotFoundError] | Failure[PathIsRootError]:
        """Get the creation date of the path.

        Args:
            path: the path to get the creation date for.

        Returns:
            The creation date of the path.
        """

        raise NotImplementedError

    @abstractmethod
    def get_all_paths(self) -> set[PureSequencePath]:
        """Get all the paths in the session.

        Returns:
            A set of all the paths in the session.
        """

        raise NotImplementedError

    @abstractmethod
    def update_creation_date(self, path: PureSequencePath, date: datetime) -> None:
        """Update the creation date of the path.

        This method is meant to be used for maintenance purposes only, such as when
        copying sequences from one session to another.

        Args:
            path: the path to update the creation date for.
            date: the new creation date.
        """

        raise NotImplementedError

    @abstractmethod
    def move(
        self, source: PureSequencePath, destination: PureSequencePath
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathExistsError]
        | Failure[PathIsSequenceError]
        | Failure[RecursivePathMoveError]
        | Failure[SequenceRunningError]
    ):
        """Move a path to a new location.

        Args:
            source: The path to move.
            destination: The new location of the path.
                If a parent of the destination path does not exist, it will be created.

        Returns:
            Success if the path was moved successfully.
            Failure with one of the following errors:

            * PathNotFoundError: If the source path does not exist.
            * PathExistsError: If the destination path already exists.
            * PathIsSequenceError: If an ancestor in the destination path is a sequence.
            * RecursivePathMoveError: If the destination path is the source path or a
                descendant of the source path.
                As a specific case, this error is returned if the source is the root
                path.
            * SequenceRunningError: If the source path contains a sequence that is
              currently running.

            If a failure is returned, the path is not moved and no path is created.
        """

        raise NotImplementedError


class PathError(RuntimeError):
    """Base class for all path errors."""

    pass


class PathNotFoundError(PathError):
    """Raised when a path is not found in the session."""

    pass


class PathIsRootError(PathError):
    """Raised when an invalid operation is performed on the root path."""

    pass


class PathHasChildrenError(PathError):
    """Raised when an invalid operation is performed on a path that has children."""

    pass


class PathExistsError(PathError):
    """Raised when a path already exists in the session."""

    pass


class RecursivePathMoveError(PathError):
    """Raised when an invalid move operation is performed."""

    pass



import contextlib
from typing import TypeVar

from caqtus.device.remote import DeviceProxy, AsyncConverter
from caqtus.shot_compilation.timed_instructions import TimedInstruction
from .runtime import Sequencer
from .trigger import Trigger
from ..remote.rpc import Proxy

SequencerType = TypeVar("SequencerType", bound=Sequencer)


class SequencerProxy(DeviceProxy[SequencerType]):
    @contextlib.asynccontextmanager
    async def program_sequence(self, sequence: TimedInstruction):
        async with self.call_method_proxy_result(
            "program_sequence", sequence
        ) as sequence_proxy:
            yield ProgrammedSequenceProxy(self.async_converter, sequence_proxy)

    async def get_trigger(self) -> Trigger:
        return await self.get_attribute("trigger")


class ProgrammedSequenceProxy:
    def __init__(self, async_converter: AsyncConverter, proxy: Proxy):
        self._async_converter = async_converter
        self._proxy = proxy

    @contextlib.asynccontextmanager
    async def run(self):
        async with (
            self._async_converter.call_method_proxy_result(
                self._proxy, "run"
            ) as run_cm_proxy,
            self._async_converter.async_context_manager(run_cm_proxy) as status_proxy,
        ):
            yield SequenceStatusProxy(self._async_converter, status_proxy)


class SequenceStatusProxy:
    def __init__(self, async_converter: AsyncConverter, proxy: Proxy):
        self._async_converter = async_converter
        self._proxy = proxy

    async def is_finished(self) -> bool:
        return await self._async_converter.call_method(self._proxy, "is_finished")



from __future__ import annotations

import abc
import datetime
from collections.abc import Mapping, Set, Iterable
from typing import Protocol, Optional, Literal, TYPE_CHECKING

import attrs

from caqtus.device import DeviceName, DeviceConfiguration
from caqtus.types.data import DataLabel, Data
from caqtus.types.iteration import IterationConfiguration, Unknown
from caqtus.types.parameter import Parameter, ParameterNamespace, ParameterSchema
from caqtus.types.timelane import TimeLanes
from caqtus.types.variable_name import DottedVariableName
from caqtus.utils.result import Success, Failure, is_failure_type, is_failure
from ._data_id import DataId
from ._exception_summary import TracebackSummary
from ._path import PureSequencePath
from ._path_hierarchy import PathError, PathNotFoundError, PathHasChildrenError
from ._shot_id import ShotId
from ._state import State

if TYPE_CHECKING:
    from ._experiment_session import ExperimentSession


class PathIsSequenceError(PathError):
    """Raised when a path is expected to be a sequence, but it is not."""

    pass


class PathIsNotSequenceError(PathError):
    """Raised when a path is expected to not be a sequence, but it is."""

    pass


class DataNotFoundError(RuntimeError):
    """Raised when data is not found in a shot."""

    pass


class SequenceStateError(RuntimeError):
    """Raised when an invalid sequence state is encountered.

    This error is raised when trying to perform an operation that is not allowed in the
    current state, such as adding data to a sequence that is not in the RUNNING state.
    """

    pass


class SequenceRunningError(SequenceStateError):
    """Raised when trying to perform an invalid operation on a running sequence."""

    pass


class SequenceNotRunningError(SequenceStateError):
    """Raised when trying to perform an invalid operation on a non-running sequence."""

    pass


class InvalidStateTransitionError(SequenceStateError):
    """Raised when an invalid state transition is attempted.

    This error is raised when trying to transition a sequence to an invalid state.
    """

    pass


class SequenceNotEditableError(SequenceStateError):
    """Raised when trying to edit a sequence that is not in the draft state."""

    pass


class SequenceNotLaunchedError(SequenceStateError):
    """Raised when accessing information only available after launching a sequence."""

    pass


class SequenceNotCrashedError(SequenceStateError):
    """Raised when trying to read the exceptions of a sequence that is not crashed."""

    pass


class ShotNotFoundError(RuntimeError):
    """Raised when a shot is not found in a sequence."""

    pass


class SequenceCollection(Protocol):
    """A collection of sequences inside a session.

    This abstract class defines the interface to read and write sequences in a session.
    Objects of this class provide methods for full access to read/write operations on
    sequences and their shots.
    However, they are not meant to be convenient to use directly in user code.
    Instead, consider using the higher-level API provided by the
    :class:`caqtus.session.Sequence` and :class:`caqtus.session.Shot` classes to access
    data from sequences and shots.
    """

    @property
    def parent_session(self) -> "ExperimentSession":
        """The session that this collection belongs to."""

        ...

    @abc.abstractmethod
    def is_sequence(
        self, path: PureSequencePath
    ) -> Success[bool] | Failure[PathNotFoundError]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_contained_sequences(
        self, path: PureSequencePath
    ) -> Success[Set[PureSequencePath]] | Failure[PathNotFoundError]:
        """Return the descendants of this path that are sequences.

        The current path is included in the result if it is a sequence.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_global_parameters(
        self, path: PureSequencePath
    ) -> (
        Success[ParameterNamespace]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotLaunchedError]
    ):
        """Get the global parameters that were used by this sequence.

        Raises:
            RuntimeError: If the sequence is in draft mode, since the global parameters
            are only set once the sequence has entered the PREPARING state.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_iteration_configuration(
        self, sequence: PureSequencePath
    ) -> IterationConfiguration:
        """Return a copy of the iteration configuration for this sequence."""

        raise NotImplementedError

    @abc.abstractmethod
    def set_iteration_configuration(
        self,
        sequence: PureSequencePath,
        iteration_configuration: IterationConfiguration,
    ) -> (
        None
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotEditableError]
    ):
        """Set the iteration configuration for this sequence."""

        raise NotImplementedError

    @abc.abstractmethod
    def get_time_lanes(
        self, sequence_path: PureSequencePath
    ) -> TimeLanes | Failure[PathNotFoundError] | Failure[PathIsNotSequenceError]:
        """Return a copy of the time lanes for this sequence."""

        raise NotImplementedError

    @abc.abstractmethod
    def set_time_lanes(
        self, sequence_path: PureSequencePath, time_lanes: TimeLanes
    ) -> (
        None
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotEditableError]
    ):
        """Set the time lanes that define how a shot is run for this sequence."""

        raise NotImplementedError

    @abc.abstractmethod
    def get_device_configurations(
        self, path: PureSequencePath
    ) -> Mapping[DeviceName, DeviceConfiguration]:
        """Get the device configurations that are used by this sequence.

        Raises:
            RuntimeError: If the sequence is in draft mode, since the device
            configurations are only set once the sequence has entered the PREPARING
            state.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def create(
        self,
        path: PureSequencePath,
        iteration_configuration: IterationConfiguration,
        time_lanes: TimeLanes,
    ) -> Success[None] | Failure[PathIsSequenceError] | Failure[PathHasChildrenError]:
        """Create a new sequence at the given path.

        Returns:
            PathIsSequenceError: If the path already exists and is a sequence.
            PathHasChildrenError: If the path already exists and has children.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_state(
        self, path: PureSequencePath
    ) -> Success[State] | Failure[PathNotFoundError] | Failure[PathIsNotSequenceError]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_exception(
        self, path: PureSequencePath
    ) -> (
        Success[Optional[TracebackSummary]]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotCrashedError]
    ):
        """Return the exceptions that occurred while running the sequence.

        Returns:
            A result wrapping the exceptions that occurred while running the sequence.

            Even if the sequence is in the CRASHED state, there may not be any
            exceptions captured. In this case, the result will be None.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def set_preparing(
        self,
        path: PureSequencePath,
        device_configurations: Mapping[DeviceName, DeviceConfiguration],
        global_parameters: ParameterNamespace,
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        """Set a sequence to the PREPARING state.

        Args:
            path: The path to the sequence to prepare.
            device_configurations: The configurations of the devices that were used to
                run this sequence.
            global_parameters: The parameters used to run the sequence.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def set_running(
        self, path: PureSequencePath, start_time: datetime.datetime | Literal["now"]
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        """Set a sequence to the RUNNING state.

        Args:
            path: The path to the sequence.
            start_time: The time at which the sequence started running.
                Must be a timezone-aware datetime object.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def set_finished(
        self, path: PureSequencePath, stop_time: datetime.datetime | Literal["now"]
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        """Set a sequence to the FINISHED state.

        Args:
            path: The path to the sequence.
            stop_time: The time at which the sequence stopped running.
                Must be a timezone-aware datetime object.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def set_interrupted(
        self, path: PureSequencePath, stop_time: datetime.datetime | Literal["now"]
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        """Set a sequence to the INTERRUPTED state.

        Args:
            path: The path to the sequence.
            stop_time: The time at which the sequence was interrupted.
                Must be a timezone-aware datetime object.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def set_crashed(
        self,
        path: PureSequencePath,
        tb_summary: TracebackSummary,
        stop_time: datetime.datetime | Literal["now"],
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        """Set the sequence to the CRASHED state.

        Args:
            path: The path of the sequence to set to the CRASHED state.
            tb_summary: A summary of the error that caused the sequence to crash.
                This summary will be saved with the sequence.
            stop_time: The time at which the sequence crashed.
                Must be a timezone-aware datetime object.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def reset_to_draft(
        self, path: PureSequencePath
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[InvalidStateTransitionError]
    ):
        """Reset the sequence to the DRAFT state.

        Warning:
            This method removes all data associated to the sequence.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_stats(
        self, path: PureSequencePath
    ) -> (
        Success[SequenceStats]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
    ):
        raise NotImplementedError

    @abc.abstractmethod
    def create_shot(
        self,
        shot_id: ShotId,
        shot_parameters: Mapping[DottedVariableName, Parameter],
        shot_data: Mapping[DataLabel, Data],
        shot_start_time: datetime.datetime,
        shot_end_time: datetime.datetime,
    ) -> (
        Success[None]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
        | Failure[SequenceNotRunningError]
    ):
        raise NotImplementedError

    @abc.abstractmethod
    def get_shots(
        self, path: PureSequencePath
    ) -> (
        Success[list[ShotId]]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
    ):
        """Return the shots that belong to this sequence."""

        raise NotImplementedError

    @abc.abstractmethod
    def get_shot_parameters(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DottedVariableName, Parameter]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_all_shot_data(
        self, path: PureSequencePath, shot_index: int
    ) -> Mapping[DataLabel, Data]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_shot_data_by_label(self, data: DataId) -> Data:
        """Return the data with the given label for the shot at the given index.

        Raises:
            PathNotFoundError: If the path does not exist in the session.
            PathIsNotSequenceError: If the path is not a sequence.
            ShotNotFoundError: If the shot does not exist in the sequence.
            DataNotFoundError: If the data with the given label does not exist in the
            shot.
        """

        raise NotImplementedError

    def get_shot_data_by_labels(
        self, path: PureSequencePath, shot_index: int, data_labels: Set[DataLabel]
    ) -> Mapping[DataLabel, Data]:
        """Return the data with the given labels for the shot at the given index.

        This method is equivalent to calling :meth:`get_shot_data_by_label` for each
        label in the set, but can be faster.

        Raises:
            PathNotFoundError: If the path does not exist in the session.
            PathIsNotSequenceError: If the path is not a sequence.
            ShotNotFoundError: If the shot does not exist in the sequence.
            DataNotFoundError: If one of the data labels does not exist in the shot.
        """

        # Naive implementation that calls get_shot_data_by_label for each label.
        shot_id = ShotId(path, shot_index)
        return {
            label: self.get_shot_data_by_label(DataId(shot_id, label))
            for label in data_labels
        }

    @abc.abstractmethod
    def get_shot_start_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime.datetime:
        raise NotImplementedError

    @abc.abstractmethod
    def get_shot_end_time(
        self, path: PureSequencePath, shot_index: int
    ) -> datetime.datetime:
        raise NotImplementedError

    @abc.abstractmethod
    def get_sequences_in_state(self, state: State) -> Iterable[PureSequencePath]:
        """Return all sequences in the given state."""

        raise NotImplementedError

    def get_parameter_schema(
        self, path: PureSequencePath
    ) -> (
        Success[ParameterSchema]
        | Failure[PathNotFoundError]
        | Failure[PathIsNotSequenceError]
    ):
        """Return the parameter schema for this sequence."""

        globals_result = self.get_global_parameters(path)
        if is_failure_type(globals_result, SequenceNotLaunchedError):
            sequence_globals = self.parent_session.get_global_parameters()
        elif is_failure(globals_result):
            return globals_result
        else:
            sequence_globals = globals_result.value

        iterations = self.get_iteration_configuration(path)

        initial_values = sequence_globals.evaluate()

        schema = iterations.get_parameter_schema(initial_values)
        return Success(schema)


@attrs.frozen
class SequenceStats:
    state: State
    start_time: Optional[datetime.datetime]
    stop_time: Optional[datetime.datetime]
    number_completed_shots: int
    expected_number_shots: int | Unknown



from collections.abc import Callable

from caqtus.utils.serialization import JSON
from ._protocol import DeviceConfigJSONSerializerProtocol
from .._configuration import DeviceConfiguration


class DeviceConfigJSONSerializer(DeviceConfigJSONSerializerProtocol):
    """Handles serialization of device configurations.

    For each device configuration type to handle, a dumper and a constructor function
    must be registered.
    """

    def __init__(self):
        self.loaders: dict[str, Callable[[JSON], DeviceConfiguration]] = {}
        self.dumpers: dict[str, Callable[[DeviceConfiguration], JSON]] = {}

    def register_device_configuration[
        C: DeviceConfiguration
    ](
        self,
        config_type: type[C],
        dumper: Callable[[C], JSON],
        constructor: Callable[[JSON], C],
    ) -> None:
        """Register a custom device configuration type for serialization.

        Args:
            config_type: A subclass of :class:`DeviceConfiguration` that is being
                registered for serialization.
            dumper: A function that will be called when it is necessary to convert a
                device configuration to JSON format.
            constructor: A function that will be called when it is necessary to build a
                device configuration from the JSON data returned by the dumper.
        """

        type_name = config_type.__qualname__

        # We need to transform the dumper into a function that can handle any device
        # configuration type, not just the one it was registered for.
        self.dumpers[type_name] = wrap_dumper(config_type, dumper)
        self.loaders[type_name] = constructor

    def dump_device_configuration(
        self, config: DeviceConfiguration
    ) -> tuple[str, JSON]:
        type_name = type(config).__qualname__
        dumper = self.dumpers[type_name]
        return type_name, dumper(config)

    def load_device_configuration(self, tag: str, content: JSON) -> DeviceConfiguration:
        constructor = self.loaders[tag]
        return constructor(content)


class wrap_dumper[C: DeviceConfiguration]:  # noqa: N801
    """Make a dumper function accept all device configuration types."""

    def __init__(self, config_type: type[C], dumper: Callable[[C], JSON]):
        self.dumper = dumper
        self.config_type = config_type

    def __call__(self, config: DeviceConfiguration) -> JSON:
        if not isinstance(config, self.config_type):
            raise TypeError(f"Expected {self.config_type}, got {type(config)}")
        return self.dumper(config)



import attrs

from caqtus.types.data import DataLabel
from ._shot_id import ShotId


@attrs.frozen
class DataId:
    """Represents a unique identifier for data in the session.

    This is used to identify data in the session and is unique for each data.

    Attributes:
        shot_id: The shot to which the data belongs.
        data_label: The label of the data.
    """

    shot_id: ShotId
    data_label: DataLabel



from typing import Protocol, TypeVar

from caqtus.device import DeviceConfiguration
from .._device_configuration_editor import DeviceConfigurationEditor

C = TypeVar("C", bound=DeviceConfiguration)


class CondetrolDeviceExtensionProtocol(Protocol):
    """Defines the operations necessary to extend Condetrol with new device."""

    def get_device_configuration_editor(
        self, device_configuration: C
    ) -> DeviceConfigurationEditor[C]:
        """Create an editor for the given device configuration.

        This method is called when the user wants to edit a device configuration.
        The returned editor will be used to display and modify the device configuration.
        """

        ...

    def available_new_configurations(self) -> set[str]:
        """Return the new configurations that can be created.

        This method is called when the user wants to create a new device configuration.
        The user will be able to choose one of the returned labels.
        """

        ...

    def create_new_device_configuration(
        self, configuration_label: str
    ) -> DeviceConfiguration:
        """Create a new device configuration.

        This method is called when the user wants to create a new device configuration.
        The label of the configuration to create is passed as an argument.
        """

        ...



from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import Optional, Any

import attrs
from PySide6.QtCore import (
    QModelIndex,
    Qt,
    QObject,
    QMimeData,
    QPersistentModelIndex,
)
from PySide6.QtGui import (
    QStandardItem,
    QStandardItemModel,
    QPalette,
    QUndoStack,
    QUndoCommand,
)

from caqtus.types.expression import Expression
from caqtus.types.iteration import (
    StepsConfiguration,
    Step,
    ExecuteShot,
    VariableDeclaration,
    LinspaceLoop,
    ArangeLoop,
)
from caqtus.types.variable_name import DottedVariableName
from caqtus.utils import serialization

NAME_COLOR = "#AA4926"
VALUE_COLOR = "#6897BB"
HIGHLIGHT_COLOR = "#cc7832"

type AnyModelIndex = QModelIndex | QPersistentModelIndex

DEFAULT_INDEX = QModelIndex()


@attrs.define
class ExecuteShotData:
    def display_data(self) -> str:
        return f"<span style='color:{HIGHLIGHT_COLOR}'>do shot</span>"


@attrs.define
class VariableDeclarationData:
    variable: DottedVariableName
    value: Expression

    def display_data(self) -> str:
        text_color = f"#{QPalette().text().color().rgba():X}"
        return (
            f"<span style='color:{NAME_COLOR}'>{self.variable}</span> "
            f"<span style='color:{text_color}'>=</span> "
            f"<span style='color:{VALUE_COLOR}'>{self.value}</span>"
        )


@attrs.define
class LinspaceLoopData:
    variable: DottedVariableName
    start: Expression
    stop: Expression
    num: int

    def display_data(self) -> str:
        text_color = f"#{QPalette().text().color().rgba():X}"
        return (
            f"<span style='color:{HIGHLIGHT_COLOR}'>for</span> "
            f"<span style='color:{NAME_COLOR}'>{self.variable}</span> "
            f"<span style='color:{text_color}'>=</span> "
            f"<span style='color:{VALUE_COLOR}'>{self.start}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>to </span> "
            f"<span style='color:{VALUE_COLOR}'>{self.stop}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>with </span> "
            f"<span style='color:{VALUE_COLOR}'>{self.num}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>steps:</span>"
        )


@attrs.define
class ArrangeLoopData:
    variable: DottedVariableName
    start: Expression
    stop: Expression
    step: Expression

    def display_data(self) -> str:
        text_color = f"#{QPalette().text().color().rgba():X}"
        return (
            f"<span style='color:{HIGHLIGHT_COLOR}'>for</span> "
            f"<span style='color:{NAME_COLOR}'>{self.variable}</span> "
            f"<span style='color:{text_color}'>=</span> "
            f"<span style='color:{VALUE_COLOR}'>{self.start}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>to </span> "
            f"<span style='color:{VALUE_COLOR}'>{self.stop}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>with </span> "
            f"<span style='color:{VALUE_COLOR}'>{self.step}</span> "
            f"<span style='color:{HIGHLIGHT_COLOR}'>spacing:</span>"
        )


StepData = (
    ExecuteShotData | VariableDeclarationData | LinspaceLoopData | ArrangeLoopData
)


class StepItem(QStandardItem):
    @classmethod
    def construct(cls, step: Step) -> StepItem:
        item = cls()
        flags = (
            Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsDragEnabled
        )
        match step:
            case ExecuteShot():
                item.setData(ExecuteShotData(), Qt.ItemDataRole.EditRole)
                item.setFlags(flags | Qt.ItemFlag.ItemNeverHasChildren)
            case VariableDeclaration(variable=variable, value=value):
                item.setData(
                    VariableDeclarationData(variable=variable, value=value),
                    Qt.ItemDataRole.EditRole,
                )
                item.setFlags(
                    flags
                    | Qt.ItemFlag.ItemIsEditable
                    | Qt.ItemFlag.ItemNeverHasChildren
                )
            case LinspaceLoop(
                variable=variable, start=start, stop=stop, num=num, sub_steps=sub_steps
            ):
                children = [cls.construct(sub_step) for sub_step in sub_steps]
                item.setData(
                    LinspaceLoopData(
                        variable=variable, start=start, stop=stop, num=num
                    ),
                    Qt.ItemDataRole.EditRole,
                )
                for child in children:
                    item.appendRow(child)
                item.setFlags(
                    flags | Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsDropEnabled
                )
            case ArangeLoop(
                variable=variable,
                start=start,
                stop=stop,
                step=loop_step,
                sub_steps=sub_steps,
            ):
                children = [cls.construct(sub_step) for sub_step in sub_steps]
                item.setData(
                    ArrangeLoopData(
                        variable=variable, start=start, stop=stop, step=loop_step
                    ),
                    Qt.ItemDataRole.EditRole,
                )
                for child in children:
                    item.appendRow(child)
                item.setFlags(
                    flags | Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsDropEnabled
                )
            case _:
                raise NotImplementedError(f"Step {step} not supported")

        return item

    def data(self, role: int = Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            step_data = self.data(Qt.ItemDataRole.EditRole)
            assert isinstance(step_data, StepData)
            return step_data.display_data()
        return super().data(role)

    def child(self, row: int, column: int | None = 0) -> StepItem:
        result = super().child(row, column)
        assert isinstance(result, StepItem)
        return result

    def get_step(self) -> Step:
        """Return a copy of the step represented by this item."""

        data = self.data(role=Qt.ItemDataRole.EditRole)
        match data:
            case ExecuteShotData():
                return ExecuteShot()
            case VariableDeclarationData(variable=variable, value=value):
                return VariableDeclaration(variable=variable, value=value)
            case LinspaceLoopData(variable=variable, start=start, stop=stop, num=num):
                child_items = [self.child(i) for i in range(self.rowCount())]
                sub_steps = [item.get_step() for item in child_items]
                return LinspaceLoop(
                    variable=variable,
                    start=start,
                    stop=stop,
                    num=num,
                    sub_steps=sub_steps,
                )
            case ArrangeLoopData(variable=variable, start=start, stop=stop, step=step):
                child_items = [self.child(i) for i in range(self.rowCount())]
                sub_steps = [item.get_step() for item in child_items]
                return ArangeLoop(
                    variable=variable,
                    start=start,
                    stop=stop,
                    step=step,
                    sub_steps=sub_steps,
                )
            case _:
                raise NotImplementedError(f"Step {data} not supported")


class StepsModel(QStandardItemModel):
    """Tree model for the steps of a sequence.

    This model holds an undo stack to allow undoing and redoing changes to the steps.
    """

    # ruff: noqa: N802

    def __init__(self, steps: StepsConfiguration, parent: Optional[QObject] = None):
        super().__init__(parent)
        self._read_only = True
        self.undo_stack = QUndoStack(self)
        self.undo_stack.setClean()
        self.set_steps(steps)

    def set_read_only(self, read_only: bool):
        self._read_only = read_only

    def is_read_only(self) -> bool:
        return self._read_only

    def get_steps(self) -> StepsConfiguration:
        root = self.invisibleRootItem()
        items = [root.child(i) for i in range(root.rowCount())]
        steps = []
        for item in items:
            assert isinstance(item, StepItem)
            steps.append(item.get_step())
        return StepsConfiguration(steps=steps)

    def set_steps(self, steps: StepsConfiguration):
        """Reset the steps contained in the model.

        This mehod clears the undo stack of the model.
        """

        self.beginResetModel()
        self.undo_stack.clear()
        self.clear()
        items = [StepItem.construct(step) for step in steps.steps]
        root = self.invisibleRootItem()
        for item in items:
            root.appendRow(item)
        self.endResetModel()

    def setData(self, index, value, role: int = Qt.ItemDataRole.EditRole):
        if self._read_only:
            return False
        previous_value = index.data(role)
        if previous_value == value:
            return False
        flat_index = into_flat_index(index)
        self.undo_stack.push(
            self.SetDataCommand(
                model=self,
                index=flat_index,
                previous_value=previous_value,
                new_value=value,
            )
        )
        return True

    @attrs.define(slots=False)
    class SetDataCommand(QUndoCommand):
        model: StepsModel
        index: FlatIndex
        previous_value: Any
        new_value: Any

        def __attrs_post_init__(self):
            super().__init__("change step value")

        def redo(self):
            index = self.model.into_index(self.index)
            QStandardItemModel.setData(
                self.model, index, self.new_value, Qt.ItemDataRole.EditRole
            )

        def undo(self):
            index = self.model.into_index(self.index)
            QStandardItemModel.setData(
                self.model, index, self.previous_value, Qt.ItemDataRole.EditRole
            )

    def flags(self, index: AnyModelIndex) -> Qt.ItemFlag:
        flags = super().flags(index)

        if self._read_only:
            flags &= ~Qt.ItemFlag.ItemIsEditable
            flags &= ~Qt.ItemFlag.ItemIsDropEnabled
        return flags

    def supportedDropActions(self) -> Qt.DropAction:
        return Qt.DropAction.MoveAction

    def mimeTypes(self) -> list[str]:
        return ["application/json"]

    def mimeData(self, indexes: Iterable[QModelIndex]) -> QMimeData:
        items = [self.itemFromIndex(index) for index in indexes]
        assert all(isinstance(item, StepItem) for item in items)
        descendants = []
        for item in items:
            descendants.extend(get_strict_descendants(item))
        items = [item for item in items if item not in descendants]

        steps = [item.get_step() for item in items]
        serialized = serialization.to_json(steps)
        mime_data = QMimeData()
        mime_data.setData("application/json", serialized.encode())
        mime_data.setText(serialized)
        return mime_data

    def itemFromIndex(self, index: AnyModelIndex) -> StepItem:
        result = super().itemFromIndex(index)
        assert isinstance(result, StepItem)
        return result

    def canDropMimeData(
        self,
        data,
        action,
        row: int,
        column: int,
        parent: QModelIndex | QPersistentModelIndex,
    ) -> bool:
        if self._read_only:
            return False
        if not data.hasFormat("application/json"):
            return False
        return super().canDropMimeData(data, action, row, column, parent)

    def dropMimeData(
        self,
        data: QMimeData,
        action: Qt.DropAction,
        row: int,
        column: int,
        parent: QModelIndex | QPersistentModelIndex,
    ) -> bool:
        if not self.canDropMimeData(data, action, row, column, parent):
            return False

        bytes_data = data.data("application/json").data()
        assert isinstance(bytes_data, bytes)
        json_string = bytes_data.decode()
        try:
            steps = serialization.from_json(json_string, list[Step])
        except ValueError:
            return False

        parent_item = (
            self.itemFromIndex(parent) if parent.isValid() else self.invisibleRootItem()
        )
        if not (parent_item.flags() & Qt.ItemFlag.ItemIsDropEnabled):
            return False
        if row == -1:
            row = parent_item.rowCount()
        self.undo_stack.push(
            self.InsertStepsCommand(
                model=self, steps=tuple(steps), parent=into_flat_index(parent), row=row
            )
        )
        return True

    @attrs.define(slots=False)
    class InsertStepsCommand(QUndoCommand):
        model: StepsModel
        steps: tuple[Step, ...]
        parent: FlatIndex
        row: int

        def __attrs_post_init__(self):
            step_text = ", ".join(f"<{step}>" for step in self.steps)
            super().__init__(f"insert {step_text}")

        def redo(self):
            parent = self.model.into_index(self.parent)
            self.model._insert_steps_without_undo(self.steps, self.row, parent)

        def undo(self):
            parent_index = self.model.into_index(self.parent)
            parent_item = (
                self.model.itemFromIndex(parent_index)
                if parent_index.isValid()
                else self.model.invisibleRootItem()
            )
            parent_item.removeRows(self.row, len(self.steps))

    def insert_steps(
        self, steps: Sequence[Step], row: int, parent: QModelIndex
    ) -> bool:
        """Insert steps at the given row under the given parent.

        This method pushes an undo command to the undo stack of the model.
        """

        if self._read_only:
            return False
        self.undo_stack.push(
            self.InsertStepsCommand(
                model=self, steps=tuple(steps), parent=into_flat_index(parent), row=row
            )
        )
        return True

    def _insert_steps_without_undo(
        self, steps: Sequence[Step], row: int, parent: QModelIndex
    ) -> bool:
        if self._read_only:
            return False
        parent_item = (
            self.itemFromIndex(parent) if parent.isValid() else self.invisibleRootItem()
        )
        new_items = [StepItem.construct(step) for step in steps]
        parent_item.insertRows(row, new_items)
        return True

    def removeRow(
        self, row, parent: QModelIndex | QPersistentModelIndex = DEFAULT_INDEX
    ) -> bool:
        return self.removeRows(row, 1, parent)

    def into_index(self, flat_index: FlatIndex) -> QModelIndex:
        index = QModelIndex()
        for row in flat_index.rows:
            index = self.index(row, 0, index)
        return index

    def removeRows(
        self, row, count, parent: QModelIndex | QPersistentModelIndex = DEFAULT_INDEX
    ) -> bool:
        if self._read_only:
            return False
        items = [
            self.itemFromIndex(self.index(row + i, 0, parent)) for i in range(count)
        ]
        steps = tuple(item.get_step() for item in items)

        self.undo_stack.push(
            self.RemoveRowsCommand(self, steps, row, into_flat_index(parent))
        )
        return True

    def remove_indices(
        self, indices: Sequence[QModelIndex | QPersistentModelIndex]
    ) -> bool:
        if self._read_only:
            return False
        # Need to be careful that the indexes are not invalidated by the removal of
        # previous rows, that's why we convert them to QPersistentModelIndex.
        persistent_indices = [
            QPersistentModelIndex(index) for index in indices if index.isValid()
        ]
        if not persistent_indices:
            return False
        self.undo_stack.beginMacro("remove steps")
        for index in persistent_indices:
            if not index.isValid():
                # It can happen that the index is no longer valid because its parent
                # was removed before. In that case, we just skip it.
                continue
            self.removeRow(index.row(), index.parent())
        self.undo_stack.endMacro()
        return True

    @attrs.define(slots=False)
    class RemoveRowsCommand(QUndoCommand):
        model: StepsModel
        steps: tuple[Step, ...]
        row: int
        parent: FlatIndex

        def __attrs_post_init__(self):
            super().__init__("remove steps")

        def redo(self):
            parent_index = self.model.into_index(self.parent)
            QStandardItemModel.removeRows(
                self.model, self.row, len(self.steps), parent_index
            )

        def undo(self):
            parent = self.model.into_index(self.parent)
            result = self.model._insert_steps_without_undo(self.steps, self.row, parent)
            assert result

    def replace_steps(self, steps: Sequence[Step], message: str) -> bool:
        if self._read_only:
            return False
        self.undo_stack.beginMacro(message)
        self.removeRows(0, self.rowCount())
        self.insert_steps(steps, 0, QModelIndex())
        self.undo_stack.endMacro()
        return True


@attrs.frozen
class FlatIndex:
    rows: tuple[int, ...]

    def parent(self) -> FlatIndex:
        if not self.rows:
            raise ValueError("Cannot get parent of root index")
        return FlatIndex(self.rows[:-1])

    def row(self) -> int:
        if not self.rows:
            raise ValueError("Cannot get row of root index")
        return self.rows[-1]


def into_flat_index(index: QModelIndex | QPersistentModelIndex) -> FlatIndex:
    rows = []
    while index.isValid():
        rows.append(index.row())
        index = index.parent()
    return FlatIndex(tuple(rows[::-1]))


def get_strict_descendants(parent: QStandardItem) -> list[QStandardItem]:
    children = [parent.child(i) for i in range(parent.rowCount())]
    descendants = []
    descendants.extend(children)
    for child in children:
        descendants.extend(get_strict_descendants(child))
    return descendants



from PySide6.QtWidgets import QLineEdit

from ._value_editor import ValueEditor


class StringEditor(ValueEditor[str]):
    """An editor to display a string.

    Args:
        placeholder: The text to display when the editor is empty.
    """

    def __init__(self, placeholder: str = "") -> None:
        self.line_edit = QLineEdit()

        if placeholder:
            self.line_edit.setPlaceholderText(placeholder)

    def set_value(self, value: str) -> None:
        self.line_edit.setText(value)

    def read_value(self) -> str:
        return self.line_edit.text()

    def set_editable(self, editable: bool) -> None:
        self.line_edit.setReadOnly(not editable)

    @property
    def widget(self) -> QLineEdit:
        return self.line_edit



from __future__ import annotations

import logging
from contextlib import AbstractContextManager
from typing import Protocol, TYPE_CHECKING

from caqtus.types.parameter import ParameterNamespace
from ._device_configuration_collection import DeviceConfigurationCollection
from ._path import PureSequencePath
from ._path_hierarchy import PathHierarchy
from ._sequence_collection import SequenceCollection

if TYPE_CHECKING:
    from ._sequence import Sequence

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class ExperimentSessionNotActiveError(RuntimeError):
    pass


class ExperimentSession(
    AbstractContextManager["ExperimentSession"],
    Protocol,
):
    """Provides a transactional connection to the storage of the experiment.

    An :py:class:`ExperimentSession` object allows to read and write configurations
    and data of the experiment.
    Every function and method that read or write data do so through an experiment
    session object.

    An experiment session object must be activated before it can be used.
    This is done by using the `with` statement on the session, inside which the session
    is active.
    If an error occurs inside the `with` block of the session, the data will be
    rolled back to the state it was in before the `with` block was entered in order to
    prevent leaving the storage in an inconsistent state.
    Data is only committed to the permanent storage when the `with` block is exited and
    will only be visible to other sessions after that point.
    For this reason, it is recommended to keep the `with` block as short as possible.

    A given session is not meant to be used concurrently.
    It can't be pickled and must not be passed to other processes.
    It is also not thread safe.
    It is not meant to be used by several coroutines at the same time, even if they
    belong to the same thread.

    It is possible to create multiple sessions connecting to the same storage using an
    :py:class:`caqtus.session.ExperimentSessionMaker`.
    """

    @property
    def paths(self) -> PathHierarchy:
        """The hierarchy of paths in the experiment session."""

        ...

    @property
    def sequences(self) -> SequenceCollection:
        """The collection of sequences in the experiment session."""

        ...

    default_device_configurations: DeviceConfigurationCollection
    """The default device configurations used to run a sequence."""

    def get_sequence(self, path: PureSequencePath | str) -> Sequence:
        """Get a sequence object from the session.

        Args:
            path: The path of the sequence to get.

        Returns:
            The sequence object.

        Raises:
            SequenceNotFoundError: If the sequence does not exist.
        """

        from ._sequence import Sequence

        return Sequence(path, self)

    def get_global_parameters(self) -> ParameterNamespace:
        """Returns a copy of the global parameters of the session.

        Global parameters are parameters that are not specific to a sequence, but are
        relevant for all the sequences.
        """

        ...

    def set_global_parameters(self, parameters: ParameterNamespace) -> None:
        """Overwrite the global parameters of the session with the given parameters."""

        ...



from __future__ import annotations

import collections
from collections.abc import Sequence, Mapping
from typing import assert_never, Optional, Any, assert_type

import attrs
import numpy as np
import numpy.typing as npt

import caqtus.formatter as fmt
from caqtus.device.output_transform import evaluate
from caqtus.types.expression import Expression
from caqtus.types.parameter import Parameters
from caqtus.types.recoverable_exceptions import InvalidValueError, InvalidTypeError
from caqtus.types.timelane import AnalogTimeLane, Ramp, Block
from caqtus.types.units import (
    ureg,
    Quantity,
    dimensionless,
    InvalidDimensionalityError,
    Unit,
    is_scalar_quantity,
)
from caqtus.types.units.base import is_in_base_units, BaseUnit
from caqtus.types.variable_name import VariableName, DottedVariableName
from ..timed_instructions import (
    TimedInstruction,
    Pattern,
    concatenate,
    create_ramp,
)
from ..timing import Time, number_ticks, start_tick, stop_tick

TIME_VARIABLE = VariableName("t")


@attrs.frozen
class DimensionedSeries[T: (np.number, np.bool_)]:
    """Represents a series of value to output on a channel with their units.

    Parameters:
        values: The sequence of values to output.
        units: The units in which the values are expressed.
            They must be in base SI units.
    """

    values: TimedInstruction[T]
    units: BaseUnit = attrs.field()

    @units.validator  # type: ignore[reportAttributeAccessIssue]
    def _validate_units(self, _, units):
        if not isinstance(units, Unit):
            raise TypeError(f"Expected a unit, got {type(units)}")
        if not is_in_base_units(units):
            raise ValueError(
                f"Unit {units} is not expressed in the base units of the registry."
            )


def compile_analog_lane(
    lane: AnalogTimeLane,
    variables: Parameters,
    step_start_times: Sequence[Time],
    time_step: Time,
) -> DimensionedSeries[np.float64]:
    """Compile the lane to a sequencer instruction.

    This function discretizes the lane time and evaluates the expressions and ramps
    in the lane for each tick.

    Args:
        lane: The lane to compile.
        variables: The values of the variables to use when evaluating the expressions
            in the lane.
        step_start_times: The start times in seconds of each step.
            This must have one more element than the number of steps in the lane, with
            the last element being the total duration of the shot.
        time_step: The time step in seconds to use for the discretization.

    Returns:
        The computed instruction for the lane.
        This contains an instruction with the values for each tick, and the unit in
        which they are expressed.
    """

    if len(lane) + 1 != len(step_start_times):
        raise ValueError(
            f"Number of steps in lane ({len(lane)}) does not match number of"
            f" step start times ({len(step_start_times)})"
        )

    # We do a first pass to compile the blocks that contain expressions, and we ignore
    # the blocks that contain ramps.
    # This is necessary because ramps need to know the value of the surrounding blocks
    # to be computed.
    expression_results: dict[Block, ConstantBlockResult | TimeDependentBlockResult] = {}
    for block_index, block_value in enumerate(lane.block_values()):
        block_start_step, block_stop_step = lane.get_block_bounds(Block(block_index))
        block_start_time = step_start_times[block_start_step]
        block_stop_time = step_start_times[block_stop_step]
        if isinstance(block_value, Expression):
            expr_result = _compile_expression_block(
                block_value,
                variables,
                block_start_time,
                block_stop_time,
                time_step,
            )
            expression_results[Block(block_index)] = expr_result
        elif isinstance(block_value, Ramp):
            continue
        else:
            assert_never(block_value)

    if len(expression_results) == 0:
        raise InvalidValueError("Lane must contain at least one expression block")

    unique_units = get_unique_units(
        {block: result.unit for block, result in expression_results.items()}
    )
    assert len(unique_units) > 0

    if len(unique_units) > 1:
        error = InvalidDimensionalityError(
            "All expressions in the lane must evaluate to the same unit"
        )
        for unit, blocks in unique_units.items():
            error.add_note(
                f"Block group {tuple(blocks)} evaluate to "
                f"{fmt.unit(unit or dimensionless)}"
            )
        raise error

    ramp_results: dict[Block, RampBlockResult] = {}
    for block_index, block_value in enumerate(lane.block_values()):
        if isinstance(block_value, Ramp):
            ramp_result = _compile_ramp_cell(
                lane,
                Block(block_index),
                expression_results,
                step_start_times,
                time_step,
            )
            ramp_results[Block(block_index)] = ramp_result
        elif isinstance(block_value, Expression):
            continue
        else:
            assert_never(block_value)

    block_results = expression_results | ramp_results

    assert len(block_results) == lane.number_blocks
    # Need to ensure that the instructions are sorted by block index before
    # concatenating
    instructions = (
        block_results[Block(block)].to_instruction()
        for block in range(lane.number_blocks)
    )
    total_instruction = concatenate(*instructions)

    units = {result.unit for result in block_results.values()}
    assert len(units) == 1
    unit = next(iter(units))

    return DimensionedSeries(total_instruction, unit)


def get_unique_units(
    units: Mapping[Block, Optional[Unit]]
) -> dict[Optional[Unit], list[Block]]:
    unique_units = collections.defaultdict(list)

    for block, unit in units.items():
        unique_units[unit].append(block)

    return unique_units


def _compile_expression_block(
    expression: Expression,
    variables: Mapping[DottedVariableName, Any],
    start_time: Time,
    stop_time: Time,
    time_step: Time,
) -> ConstantBlockResult | TimeDependentBlockResult:
    if is_constant(expression):
        length = number_ticks(start_time, stop_time, time_step)
        return evaluate_constant_expression(expression, variables, length)
    else:
        return evaluate_time_dependent_expression(
            expression, variables, start_time, stop_time, time_step
        )


def evaluate_constant_expression(
    expression: Expression,
    variables: Mapping[DottedVariableName, Any],
    length: int,
) -> ConstantBlockResult:
    value = evaluate(expression, variables)

    if is_scalar_quantity(value):
        return ConstantBlockResult(
            value=value.magnitude,
            length=length,
            unit=value.units,
        )
    else:
        assert_type(value, float | int | bool)
        return ConstantBlockResult(
            value=float(value),
            length=length,
            unit=dimensionless,
        )


def evaluate_time_dependent_expression(
    expression: Expression,
    variables: Mapping[DottedVariableName, Any],
    start_time: Time,
    stop_time: Time,
    time_step: Time,
) -> TimeDependentBlockResult:
    assert not is_constant(expression)

    time_values = get_time_array(start_time, stop_time, time_step) - float(start_time)
    # The first time is not necessarily 0, it is the time of the first tick of the
    # block.
    # Same for the last time which is not necessarily stop_time - start_time, but the
    # time of the last tick of the block.
    # If a ramp precedes this block, we also want to know the value of the current block
    # at true t=0, so we compute this as well.
    # Same if a ramp follows this block, we want to know the value of the current block
    # at true t=stop_time - start_time.
    time_values = np.insert(
        time_values, [0, len(time_values)], [0, float(stop_time - start_time)]
    )

    t = time_values * ureg.s
    variables = dict(variables) | {TIME_VARIABLE: t}
    evaluated = expression.evaluate(variables)

    if isinstance(evaluated, Quantity):
        in_base_units = evaluated.to_base_units()
        magnitudes = in_base_units.magnitude
        if not isinstance(magnitudes, np.ndarray):
            raise InvalidTypeError(
                f"{fmt.expression(expression)} does not evaluate to a series of values"
            )
        unit = in_base_units.units
    elif isinstance(evaluated, np.ndarray):
        magnitudes = evaluated
        unit = dimensionless
    else:
        raise InvalidTypeError(
            f"{fmt.expression(expression)} does not evaluate to a series of values"
        )
    length = number_ticks(start_time, stop_time, time_step)
    if magnitudes.shape != (length + 2,):
        raise InvalidValueError(
            f"{fmt.expression(expression)} evaluates to an array of shape"
            f" {magnitudes.shape} while a shape of {(length,)} is expected",
        )
    return TimeDependentBlockResult(
        values=magnitudes[1:-1].astype(np.float64),
        unit=unit,
        initial_value=float(magnitudes[0]),
        final_value=float(magnitudes[-1]),
    )


def _compile_ramp_cell(
    lane: AnalogTimeLane,
    ramp_block: Block,
    expression_blocks: dict[Block, ConstantBlockResult | TimeDependentBlockResult],
    step_bounds: Sequence[Time],
    time_step: Time,
) -> RampBlockResult:
    previous_block = Block(ramp_block - 1)
    if previous_block < 0:
        raise InvalidValueError("There can't be a ramp at the beginning of a lane")
    next_block = Block(ramp_block + 1)
    if next_block >= lane.number_blocks:
        raise InvalidValueError("There can't be a ramp at the end of a lane")

    try:
        previous_block_result = expression_blocks[previous_block]
    except KeyError:
        raise InvalidValueError(
            f"Block {previous_block} that precedes a ramp must be an expression"
        ) from None
    try:
        next_block_result = expression_blocks[next_block]
    except KeyError:
        raise InvalidValueError(
            f"Block {next_block} that follows a ramp must be an expression"
        ) from None

    assert previous_block_result.unit == next_block_result.unit

    ramp_start_value = previous_block_result.get_final_value()
    ramp_end_value = next_block_result.get_initial_value()

    ramp_start_step, ramp_end_step = lane.get_block_bounds(ramp_block)

    ramp_start_time = step_bounds[ramp_start_step]
    ramp_end_time = step_bounds[ramp_end_step]

    return RampBlockResult.through_two_points(
        ramp_start_time,
        ramp_start_value,
        ramp_end_time,
        ramp_end_value,
        time_step,
        previous_block_result.unit,
    )


def is_constant(expression: Expression) -> bool:
    return TIME_VARIABLE not in expression.upstream_variables


def get_time_array(
    start: Time, stop: Time, time_step: Time
) -> np.ndarray[Any, np.dtype[np.floating]]:
    times = np.arange(
        start_tick(start, time_step),
        stop_tick(stop, time_step),
        dtype=np.float64,
    ) * float(time_step)
    return times


@attrs.frozen
class ConstantBlockResult:
    """Result of compiling a constant block."""

    value: float
    length: int
    unit: BaseUnit

    def get_initial_value(self) -> float:
        return self.value

    def get_final_value(self) -> float:
        return self.value

    def to_instruction(self) -> TimedInstruction[np.float64]:
        return Pattern([self.value]) * self.length


@attrs.frozen(eq=False)
class TimeDependentBlockResult:
    """Result of compiling a time-dependent block."""

    values: npt.NDArray[np.float64]
    unit: BaseUnit

    initial_value: float
    final_value: float

    def __eq__(self, other):
        if not isinstance(other, TimeDependentBlockResult):
            return NotImplemented
        return (
            np.allclose(self.values, other.values)
            and self.unit == other.unit
            and self.initial_value == other.initial_value
            and self.final_value == other.final_value
        )

    def get_initial_value(self) -> float:
        return self.initial_value

    def get_final_value(self) -> float:
        return self.final_value

    def to_instruction(self) -> TimedInstruction[np.float64]:
        return Pattern(self.values, dtype=np.dtype(np.float64))


@attrs.frozen
class RampBlockResult:
    """Result of compiling a ramp block."""

    initial_value: float
    final_value: float
    length: int

    unit: BaseUnit

    @classmethod
    def through_two_points(
        cls,
        t0: Time,
        v0: float,
        t1: Time,
        v1: float,
        time_step: Time,
        unit: BaseUnit,
    ) -> RampBlockResult:
        def f(t: Time) -> float:
            return float((t - t0) / (t1 - t0)) * (v1 - v0) + v0

        first_tick = start_tick(t0, time_step)
        last_tick = stop_tick(t1, time_step)

        first_tick_time = Time(first_tick * time_step)
        last_tick_time = Time(last_tick * time_step)

        length = last_tick - first_tick

        if length == 0:
            # We can pick whatever value we want, as the ramp has no length.
            # However, we need to be careful to not call f with t0 == t1 as this will
            # raise a ZeroDivisionError.
            initial_value = v0
            final_value = v1
        else:
            initial_value = f(first_tick_time)
            final_value = f(last_tick_time)

        return RampBlockResult(
            initial_value,
            final_value,
            length,
            unit,
        )

    def to_instruction(self) -> TimedInstruction[np.float64]:
        return create_ramp(self.initial_value, self.final_value, self.length)



from __future__ import annotations

import enum
import functools
from collections.abc import Callable

from PySide6 import QtWidgets

from ._editor_builder import ValueEditor


def generate_enum_editor[
    T: enum.Enum
](enum_type: type[T]) -> Callable[[], EnumEditor[T]]:
    """Generate an editor factory for an enum type.

    Returns:
        A factory function that creates an EnumEditor for the given enum type.
        The editor is a combobox with the string representation of each enum member as
        its items.
    """

    return functools.partial(EnumEditor[T], enum_type)


class EnumEditor[T: enum.Enum](ValueEditor[T]):
    def __init__(self, enum_type: type[T]):
        self._enum_type = enum_type
        self.combobox = QtWidgets.QComboBox()
        self._index_to_value = {}
        self._value_to_index = {}
        for index, value in enumerate(enum_type):
            self.combobox.addItem(str(value))
            self._index_to_value[index] = value
            self._value_to_index[value] = index

    def set_value(self, value: T) -> None:
        try:
            index = self._value_to_index[value]
        except KeyError:
            raise ValueError(
                f"Value {value} is not a member of the enum {self._enum_type}"
            ) from None
        self.combobox.setCurrentIndex(index)

    def read_value(self) -> T:
        index = self.combobox.currentIndex()
        return self._index_to_value[index]

    def set_editable(self, editable: bool) -> None:
        self.combobox.setEnabled(editable)

    @property
    def widget(self) -> QtWidgets.QComboBox:
        return self.combobox
